from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError

from .dependency import DependencyDefinition, MultiDependencyDefinition, SolidInvocation
from .output import OutputDefinition
from .solid import ISolidDefinition

_composition_stack = []


def enter_composition(name, source):
    _composition_stack.append(InProgressCompositionContext(name, source))


def exit_composition(output=None):
    return _composition_stack.pop().complete(output)


def current_context():
    return _composition_stack[-1]


def assert_in_composition(solid_name):
    if len(_composition_stack) < 1:
        raise DagsterInvariantViolationError(
            'Attempted to call solid "{solid_name}" outside of a composition function. '
            'Calling solids is only valid in a function decorated with '
            '@pipeline or @composite_solid.'.format(solid_name=solid_name)
        )


class InProgressCompositionContext:
    '''This context captures invocations of solids within a
    composition function such as @composite_solid or @pipeline
    '''

    def __init__(self, name, source):
        self.name = check.str_param(name, 'name')
        self.source = check.str_param(source, 'source')
        self._invocations = {}
        self._collisions = {}

    def observe_invocation(self, given_alias, solid_def, input_bindings, input_mappings):

        if given_alias is None:
            solid_name = solid_def.name
            if self._collisions.get(solid_name):
                self._collisions[solid_name] += 1
                solid_name = '{solid_name}_{n}'.format(
                    solid_name=solid_name, n=self._collisions[solid_name]
                )
            else:
                self._collisions[solid_name] = 1
        else:
            solid_name = given_alias

        if self._invocations.get(solid_name):
            raise DagsterInvalidDefinitionError(
                '{source} {name} invoked the same solid ({solid_name}) twice without aliasing.'.format(
                    source=self.source, name=self.name, solid_name=solid_name
                )
            )

        self._invocations[solid_name] = InvokedSolidNode(
            solid_name, solid_def, input_bindings, input_mappings
        )
        return solid_name

    def complete(self, output):
        return CompleteCompositionContext(
            self.name, self._invocations, check.opt_dict_param(output, 'output')
        )


class CompleteCompositionContext(
    namedtuple(
        '_CompositionContext', 'name solid_defs dependencies input_mappings output_mapping_dict'
    )
):
    '''The processed information from capturing solid invocations during a composition function.
    '''

    def __new__(cls, name, invocations, output_mapping_dict):

        dep_dict = {}
        solid_def_dict = {}
        input_mappings = []

        for invocation in invocations.values():
            def_name = invocation.solid_def.name
            if def_name in solid_def_dict and solid_def_dict[def_name] is not invocation.solid_def:
                check.failed(
                    'Detected conflicting solid definitions with the same name "{name}"'.format(
                        name=def_name
                    )
                )
            solid_def_dict[def_name] = invocation.solid_def

            deps = {}
            for input_name, node in invocation.input_bindings.items():
                if isinstance(node, InvokedSolidOutputHandle):
                    deps[input_name] = DependencyDefinition(node.solid_name, node.output_name)
                elif isinstance(node, list) and all(
                    map(lambda item: isinstance(item, InvokedSolidOutputHandle), node)
                ):
                    deps[input_name] = MultiDependencyDefinition(
                        [DependencyDefinition(call.solid_name, call.output_name) for call in node]
                    )
                else:
                    check.failed('Unexpected input binding - got {node}'.format(node=node))

            dep_dict[SolidInvocation(invocation.solid_def.name, invocation.solid_name)] = deps

            for input_name, node in invocation.input_mappings.items():
                input_mappings.append(node.input_def.mapping_to(invocation.solid_name, input_name))

        return super(cls, CompleteCompositionContext).__new__(
            cls, name, list(solid_def_dict.values()), dep_dict, input_mappings, output_mapping_dict
        )


class CallableSolidNode:
    '''An intermediate object in solid composition to allow for binding information such as
    an alias before invoking.
    '''

    def __init__(self, solid_def, given_alias=None):
        self.solid_def = solid_def
        self.given_alias = check.opt_str_param(given_alias, 'given_alias')

    def __call__(self, *args, **kwargs):
        solid_name = self.given_alias if self.given_alias else self.solid_def.name
        assert_in_composition(solid_name)

        input_bindings = {}
        input_mappings = {}

        # handle *args
        for idx, output_node in enumerate(args):
            if idx >= len(self.solid_def.input_defs):
                raise DagsterInvalidDefinitionError(
                    'In {source} {name} received too many inputs for solid '
                    'invocation {solid_name}. Only {def_num} defined, received {arg_num}'.format(
                        source=current_context().source,
                        name=current_context().name,
                        solid_name=solid_name,
                        def_num=len(self.solid_def.input_defs),
                        arg_num=len(args),
                    )
                )

            input_name = self.solid_def.input_defs[idx].name

            self._process_argument_node(
                solid_name,
                output_node,
                input_name,
                input_mappings,
                input_bindings,
                '(at position {idx})'.format(idx=idx),
            )

        # then **kwargs
        for input_name, output_node in kwargs.items():
            self._process_argument_node(
                solid_name,
                output_node,
                input_name,
                input_mappings,
                input_bindings,
                '(passed by keyword)',
            )

        solid_name = current_context().observe_invocation(
            self.given_alias, self.solid_def, input_bindings, input_mappings
        )

        if len(self.solid_def.output_defs) == 0:
            return None

        if len(self.solid_def.output_defs) == 1:
            output_name = self.solid_def.output_defs[0].name
            return InvokedSolidOutputHandle(solid_name, output_name)

        outputs = [output_def.name for output_def in self.solid_def.output_defs]
        return namedtuple('_{solid_def}_outputs'.format(solid_def=self.solid_def.name), outputs)(
            **{output: InvokedSolidOutputHandle(solid_name, output) for output in outputs}
        )

    def _process_argument_node(
        self, solid_name, output_node, input_name, input_mappings, input_bindings, arg_desc
    ):

        if isinstance(output_node, InvokedSolidOutputHandle):
            input_bindings[input_name] = output_node
        elif isinstance(output_node, InputMappingNode):
            input_mappings[input_name] = output_node
        elif isinstance(output_node, list):
            if all(map(lambda item: isinstance(item, InvokedSolidOutputHandle), output_node)):
                input_bindings[input_name] = output_node

            else:
                raise DagsterInvalidDefinitionError(
                    'In {source} {name} received a list containing invalid types for input '
                    '"{input_name}" {arg_desc} in solid invocation {solid_name}. '
                    'Lists can only contain the output from previous solid invocations.'.format(
                        source=current_context().source,
                        name=current_context().name,
                        arg_desc=arg_desc,
                        input_name=input_name,
                        solid_name=solid_name,
                    )
                )

        elif isinstance(output_node, tuple) and all(
            map(lambda item: isinstance(item, InvokedSolidOutputHandle), output_node)
        ):
            raise DagsterInvalidDefinitionError(
                'In {source} {name} received a tuple of multiple outputs for '
                'input "{input_name}" {arg_desc} in solid invocation {solid_name}. '
                'Must pass individual output, available from tuple: {options}'.format(
                    source=current_context().source,
                    name=current_context().name,
                    arg_desc=arg_desc,
                    input_name=input_name,
                    solid_name=solid_name,
                    options=output_node._fields,
                )
            )

        else:
            raise DagsterInvalidDefinitionError(
                'In {source} {name} received invalid type {type} for input '
                '"{input_name}" {arg_desc} in solid invocation "{solid_name}". '
                'Must pass the output from previous solid invocations or inputs to the '
                'composition function as inputs when invoking solids during composition.'.format(
                    source=current_context().source,
                    name=current_context().name,
                    type=type(output_node),
                    arg_desc=arg_desc,
                    input_name=input_name,
                    solid_name=solid_name,
                )
            )


class InvokedSolidNode(
    namedtuple('_InvokedSolidNode', 'solid_name solid_def input_bindings input_mappings')
):
    '''The metadata about a solid invocation saved by the current composition context.
    '''

    def __new__(cls, solid_name, solid_def, input_bindings, input_mappings):
        return super(cls, InvokedSolidNode).__new__(
            cls,
            check.str_param(solid_name, 'solid_name'),
            check.inst_param(solid_def, 'solid_def', ISolidDefinition),
            check.dict_param(input_bindings, 'input_bindings', key_type=str),
            check.dict_param(
                input_mappings, 'input_mappings', key_type=str, value_type=InputMappingNode
            ),
        )


class InvokedSolidOutputHandle(object):
    '''The return value for an output when invoking a solid in a composition function.
    '''

    def __init__(self, solid_name, output_name):
        self.solid_name = check.str_param(solid_name, 'solid_name')
        self.output_name = check.str_param(output_name, 'output_name')

    def __iter__(self):
        raise DagsterInvariantViolationError(
            'Attempted to iterate over an {cls}. This object represents the output "{out}" '
            'from the solid "{solid}". Consider yielding multiple Outputs if you seek to pass '
            'different parts of this output to different solids.'.format(
                cls=self.__class__.__name__, out=self.output_name, solid=self.solid_name
            )
        )

    def __getitem__(self, idx):
        raise DagsterInvariantViolationError(
            'Attempted to index in to an {cls}. This object represents the output "{out}" '
            'from the solid "{solid}". Consider yielding multiple Outputs if you seek to pass '
            'different parts of this output to different solids.'.format(
                cls=self.__class__.__name__, out=self.output_name, solid=self.solid_name
            )
        )


class InputMappingNode:
    def __init__(self, input_def):
        self.input_def = input_def


def composite_mapping_from_output(output, output_defs, solid_name):
    # output can be different types
    check.list_param(output_defs, 'output_defs', OutputDefinition)
    check.str_param(solid_name, 'solid_name')

    # single output
    if isinstance(output, InvokedSolidOutputHandle):
        if len(output_defs) == 1:
            defn = output_defs[0]
            return {defn.name: defn.mapping_from(output.solid_name, output.output_name)}
        else:
            raise DagsterInvalidDefinitionError(
                'Returned a single output ({solid_name}.{output_name}) in '
                '@composite_solid {name} but {num} outputs are defined. '
                'Return a dict to map defined outputs.'.format(
                    solid_name=output.solid_name,
                    output_name=output.output_name,
                    name=solid_name,
                    num=len(output_defs),
                )
            )

    output_mapping_dict = {}
    output_def_dict = {output_def.name: output_def for output_def in output_defs}

    # tuple returned directly
    if isinstance(output, tuple) and all(
        map(lambda item: isinstance(item, InvokedSolidOutputHandle), output)
    ):
        for handle in output:
            if handle.output_name not in output_def_dict:
                raise DagsterInvalidDefinitionError(
                    'Output name mismatch returning output tuple in @composite_solid {name}. '
                    'No matching OutputDefinition named {output_name} for {solid_name}.{output_name}.'
                    'Return a dict to map to the desired OutputDefinition'.format(
                        name=solid_name,
                        output_name=handle.output_name,
                        solid_name=handle.solid_name,
                    )
                )
            output_mapping_dict[handle.output_name] = output_def_dict[
                handle.output_name
            ].mapping_from(handle.solid_name, handle.output_name)

        return output_mapping_dict

    # mapping dict
    if isinstance(output, dict):
        for name, handle in output.items():
            if name not in output_def_dict:
                raise DagsterInvalidDefinitionError(
                    '@composite_solid {name} referenced key {key} which does not match any '
                    'OutputDefinitions. Valid options are: {options}'.format(
                        name=solid_name, key=name, options=list(output_def_dict.keys())
                    )
                )
            if not isinstance(handle, InvokedSolidOutputHandle):
                raise DagsterInvalidDefinitionError(
                    '@composite_solid {name} returned problematic dict entry under '
                    'key {key} of type {type}. Dict values must be outputs of '
                    'invoked solids'.format(name=solid_name, key=name, type=type(handle))
                )

            output_mapping_dict[name] = output_def_dict[name].mapping_from(
                handle.solid_name, handle.output_name
            )

        return output_mapping_dict

    # error
    if output is not None:
        raise DagsterInvalidDefinitionError(
            '@composite_solid {name} returned problematic value '
            'of type {type}. Expected return value from invoked solid or dict mapping '
            'output name to return values from invoked solids'.format(
                name=solid_name, type=type(output)
            )
        )



from dagster import check
from dagster.core.storage.file_manager import FileManager
from dagster.core.storage.intermediates_manager import IntermediatesManager
from dagster.core.types.field_utils import check_user_facing_opt_field_param

from .config import resolve_config_field


class SystemStorageDefinition:
    '''
    Dagster stores run metadata and intermediate data on the user's behalf.
    The SystemStorageDefinition exists in order to configure and customize
    those behaviors.

    Example storage definitions are the mem_system_storage in this module,
    which stores all intermediates and run data in memory. and the fs_system_storage,
    which stores all that data in the local filesystem.

    In dagster_aws there is the S3SystemStorageDefinition. We anticipate having
    system storage for every major cloud provider. And it is user customizable
    for users with custom infrastructure needs.

    The storage definitions passed into the ModeDefinition determine the config
    schema of the "storage" section of the environment configuration.

    Args:
        name (str): Name of the storage mode.
        is_persistent (bool): Does storage def persist in a way that can cross process/node
            boundaries. Execution with, for example, the multiprocess executor or within
            the context of dagster-airflow requires a persistent storage mode.
        config_field (Field): Configuration field for its section of the storage config.
        system_storage_creation_fn: (Callable[InitSystemStorageContext, SystemStorageData])
            Called by the system. The author of the StorageSystemDefinition must provide this function,
            which consumes the init context and then emits the SystemStorageData.
        required_resource_keys(Set[str]):
            The resources that this storage needs at runtime to function.
    '''

    def __init__(
        self,
        name,
        is_persistent,
        config_field=None,
        system_storage_creation_fn=None,
        required_resource_keys=None,
    ):
        self.name = check.str_param(name, 'name')
        self.is_persistent = check.bool_param(is_persistent, 'is_persistent')
        self.config_field = check_user_facing_opt_field_param(
            config_field,
            'config_field',
            'of a SystemStorageDefinition named {name}'.format(name=self.name),
        )
        self.system_storage_creation_fn = check.opt_callable_param(
            system_storage_creation_fn, 'system_storage_creation_fn'
        )
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, 'required_resource_keys', of_type=str
        )


class SystemStorageData:
    def __init__(self, intermediates_manager, file_manager):
        self.intermediates_manager = check.inst_param(
            intermediates_manager, 'intermediates_manager', IntermediatesManager
        )
        self.file_manager = check.inst_param(file_manager, 'file_manager', FileManager)


def system_storage(
    name=None, is_persistent=True, config_field=None, config=None, required_resource_keys=None
):
    '''A decorator for creating a SystemStorageDefinition. The decorated function will be used as the
    system_storage_creation_fn in a SystemStorageDefinition.

    Args:
        name (str)
        is_persistent (bool): Does storage def persist in way that can cross process/node
            boundaries. Execution with, for example, the multiprocess executor or within
            the context of dagster-airflow require a persistent storage mode.
        required_resource_keys (Set[str]):
            The resources that this storage needs at runtime to function.
        config (Dict[str, Field]):
            The schema for the configuration data made available to the system_storage_creation_fn.
        config_field (Field):
            Used in the rare case of a top level config type other than a dictionary.

            Only one of config or config_field can be provided.

    '''

    if callable(name):
        check.invariant(is_persistent is True)
        check.invariant(config_field is None)
        check.invariant(config is None)
        check.invariant(required_resource_keys is None)
        return _SystemStorageDecoratorCallable()(name)

    return _SystemStorageDecoratorCallable(
        name=name,
        is_persistent=is_persistent,
        config_field=resolve_config_field(config_field, config, '@system_storage'),
        required_resource_keys=required_resource_keys,
    )


class _SystemStorageDecoratorCallable:
    def __init__(
        self, name=None, is_persistent=True, config_field=None, required_resource_keys=None
    ):
        self.name = check.opt_str_param(name, 'name')
        self.is_persistent = check.bool_param(is_persistent, 'is_persistent')
        self.config_field = config_field  # type check in definition
        self.required_resource_keys = required_resource_keys  # type check in definition

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        return SystemStorageDefinition(
            name=self.name,
            is_persistent=self.is_persistent,
            config_field=self.config_field,
            system_storage_creation_fn=fn,
            required_resource_keys=self.required_resource_keys,
        )



import atexit
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from enum import Enum

import six
from rx import Observable

from dagster import check

MAX_BYTES_FILE_READ = 33554432  # 32 MB
MAX_BYTES_CHUNK_READ = 4194304  # 4 MB


class ComputeIOType(Enum):
    STDOUT = 'stdout'
    STDERR = 'stderr'


ComputeLogFileData = namedtuple('ComputeLogFileData', 'path data cursor size download_url')


class ComputeLogManager(six.with_metaclass(ABCMeta)):
    # API
    @abstractmethod
    def get_local_path(self, run_id, step_key, io_type):
        pass

    @abstractmethod
    def is_compute_completed(self, run_id, step_key):
        pass

    @abstractmethod
    def on_compute_start(self, step_context):
        pass

    @abstractmethod
    def on_compute_finish(self, step_context):
        pass

    @abstractmethod
    def download_url(self, run_id, step_key, io_type):
        pass

    @abstractmethod
    def read_logs_file(self, run_id, step_key, io_type, cursor=0, max_bytes=MAX_BYTES_FILE_READ):
        pass

    @abstractmethod
    def on_subscribe(self, subscription):
        pass

    def enabled(self, _step_context):
        return True

    def observable(self, run_id, step_key, io_type, cursor=None):
        check.str_param(run_id, 'run_id')
        check.str_param(step_key, 'step_key')
        check.inst_param(io_type, 'io_type', ComputeIOType)
        check.opt_str_param(cursor, 'cursor')

        if cursor:
            cursor = int(cursor)
        else:
            cursor = 0

        subscription = ComputeLogSubscription(self, run_id, step_key, io_type, cursor)
        self.on_subscribe(subscription)
        return Observable.create(subscription)  # pylint: disable=E1101


class ComputeLogSubscription(object):
    def __init__(self, manager, run_id, step_key, io_type, cursor):
        self.manager = manager
        self.run_id = run_id
        self.step_key = step_key
        self.io_type = io_type
        self.cursor = cursor
        self.observer = None
        atexit.register(self._clean)

    def __call__(self, observer):
        self.observer = observer
        self.fetch()
        if self.manager.is_compute_completed(self.run_id, self.step_key):
            self.complete()

    def fetch(self):
        if not self.observer:
            return

        should_fetch = True
        while should_fetch:
            update = self.manager.read_logs_file(
                self.run_id,
                self.step_key,
                self.io_type,
                self.cursor,
                max_bytes=MAX_BYTES_CHUNK_READ,
            )
            if not self.cursor or update.cursor != self.cursor:
                self.observer.on_next(update)
                self.cursor = update.cursor
            should_fetch = update.data and len(update.data.encode('utf-8')) >= MAX_BYTES_CHUNK_READ

    def complete(self):
        if not self.observer:
            return
        self.observer.on_completed()

    def _clean(self):
        self.complete()
        self.observer = None



import pytest

from dagster import (
    Bool,
    DagsterExecutionStepExecutionError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    Dict,
    Field,
    ModeDefinition,
    PipelineDefinition,
    PresetDefinition,
    execute_pipeline_with_preset,
    lambda_solid,
    solid,
)
from dagster.utils import script_relative_path


def test_presets():
    @solid(config_field=Field(Dict(fields={'error': Field(Bool)})))
    def can_fail(context):
        if context.solid_config['error']:
            raise Exception('I did an error')
        return 'cool'

    @lambda_solid
    def always_fail():
        raise Exception('I always do this')

    pipeline = PipelineDefinition(
        name='simple',
        solid_defs=[can_fail, always_fail],
        preset_defs=[
            PresetDefinition.from_files(
                'passing',
                environment_files=[script_relative_path('pass_env.yaml')],
                solid_subset=['can_fail'],
            ),
            PresetDefinition(
                'passing_direct_dict',
                environment_dict={'solids': {'can_fail': {'config': {'error': False}}}},
                solid_subset=['can_fail'],
            ),
            PresetDefinition.from_files(
                'failing_1',
                environment_files=[script_relative_path('fail_env.yaml')],
                solid_subset=['can_fail'],
            ),
            PresetDefinition.from_files(
                'failing_2', environment_files=[script_relative_path('pass_env.yaml')]
            ),
        ],
    )

    with pytest.raises(DagsterInvalidDefinitionError):
        PresetDefinition.from_files(
            'invalid_1', environment_files=[script_relative_path('not_a_file.yaml')]
        )

    with pytest.raises(DagsterInvariantViolationError):
        PresetDefinition.from_files(
            'invalid_2', environment_files=[script_relative_path('test_repository_definition.py')]
        )

    assert execute_pipeline_with_preset(pipeline, 'passing').success

    assert execute_pipeline_with_preset(pipeline, 'passing_direct_dict').success

    with pytest.raises(DagsterExecutionStepExecutionError):
        execute_pipeline_with_preset(pipeline, 'failing_1')

    with pytest.raises(DagsterExecutionStepExecutionError):
        execute_pipeline_with_preset(pipeline, 'failing_2')

    with pytest.raises(DagsterInvariantViolationError, match="Could not find preset"):
        execute_pipeline_with_preset(pipeline, 'not_failing')


def test_invalid_preset():
    @lambda_solid
    def lil_solid():
        return ';)'

    with pytest.raises(DagsterInvalidDefinitionError, match='mode "mode_b" which is not defined'):
        PipelineDefinition(
            name="preset_modes",
            solid_defs=[lil_solid],
            mode_defs=[ModeDefinition(name="mode_a")],
            preset_defs=[PresetDefinition(name="preset_b", mode="mode_b")],
        )



'''Internal py2/3 compatibility library. A little more than six.'''

import inspect
import os
import signal
import sys
import time

from .json import JSONDecodeError, dump, dumps
from .temp_dir import get_system_temp_directory

try:
    # Python 2 tempfile doesn't have tempfile.TemporaryDirectory
    import backports.tempfile as tempfile
except ImportError:
    import tempfile


TemporaryDirectory = tempfile.TemporaryDirectory

try:
    # pylint:disable=redefined-builtin,self-assigning-variable
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = IOError


try:
    # pylint:disable=redefined-builtin,self-assigning-variable
    ModuleNotFoundError = ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError

try:
    import _thread as thread
except ImportError:
    import thread

IS_WINDOWS = os.name == 'nt'

# TODO implement a generic import by name -- see https://stackoverflow.com/questions/301134/how-to-import-a-module-given-its-name

# https://stackoverflow.com/a/67692/324449
def import_module_from_path(module_name, path_to_file):
    version = sys.version_info
    if version.major >= 3 and version.minor >= 5:
        import importlib.util

        spec = importlib.util.spec_from_file_location(module_name, path_to_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
    elif version.major >= 3 and version.minor >= 3:
        from importlib.machinery import SourceFileLoader

        # pylint:disable=deprecated-method, no-value-for-parameter
        module = SourceFileLoader(module_name, path_to_file).load_module()
    else:
        from imp import load_source

        module = load_source(module_name, path_to_file)

    return module


def is_ascii(str_):
    if sys.version_info.major < 3:
        try:
            str_.decode('ascii')
            return True
        except UnicodeEncodeError:
            return False
    elif sys.version_info.major == 3 and sys.version_info.minor < 7:
        try:
            str_.encode('ascii')
            return True
        except UnicodeEncodeError:
            return False
    else:
        return str_.isascii()


if sys.version_info.major >= 3 and sys.version_info.minor >= 3:
    time_fn = time.perf_counter
elif IS_WINDOWS:
    time_fn = time.clock
else:
    time_fn = time.time

try:
    from unittest import mock
except ImportError:
    # Because this dependency is not encoded setup.py deliberately
    # (we do not want to override or conflict with our users mocks)
    # we never fail when importing this.

    # This will only be used within *our* test enviroment of which
    # we have total control
    try:
        import mock
    except ImportError:
        pass


def get_args(callble):
    if sys.version_info.major >= 3:
        return [
            parameter.name
            for parameter in inspect.signature(callble).parameters.values()
            if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        ]
    else:
        try:
            if inspect.isclass(callble):
                return inspect.getargspec(callble.__init__)  # pylint: disable=deprecated-method
            return inspect.getargspec(callble)  # pylint: disable=deprecated-method
        except TypeError:
            # This will happen when we try to get the argspec for a slot wrapper, e.g.:
            # TypeError: <slot wrapper '__init__' of 'object' objects> is not a Python function
            return None



from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.serdes import whitelist_for_serdes


@whitelist_for_serdes
class ScheduleDefinitionData(
    namedtuple('ScheduleDefinitionData', 'name cron_schedule environment_vars')
):
    def __new__(cls, name, cron_schedule, environment_vars=None):
        return super(ScheduleDefinitionData, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(cron_schedule, 'cron_schedule'),
            check.opt_dict_param(environment_vars, 'environment_vars'),
        )


class ScheduleDefinition(object):
    '''Define a schedule that targets a pipeline

    Args:
        name (str): The name of the schedule.
        cron_schedule (str): A valid cron string for the schedule
        environment_dict (Optional[dict]): The enviroment configuration that parameterizes this
            execution, as a dict.
        should_execute (Optional[function]): Function that runs at schedule execution time that
            determines whether a schedule should execute. Defaults to a function that always returns
            ``True``.
        environment_vars (Optional[dict]): The environment variables to set for the schedule
    '''

    __slots__ = [
        '_schedule_definition_data',
        '_execution_params',
        '_environment_dict_fn',
        '_should_execute',
    ]

    def __init__(
        self,
        name,
        cron_schedule,
        pipeline_name,
        environment_dict=None,
        environment_dict_fn=None,
        tags=None,
        mode="default",
        should_execute=lambda: True,
        environment_vars=None,
    ):
        self._schedule_definition_data = ScheduleDefinitionData(
            name=check.str_param(name, 'name'),
            cron_schedule=check.str_param(cron_schedule, 'cron_schedule'),
            environment_vars=check.opt_dict_param(environment_vars, 'environment_vars'),
        )

        check.str_param(pipeline_name, 'pipeline_name')
        environment_dict = check.opt_dict_param(environment_dict, 'environment_dict')
        tags = check.opt_list_param(tags, 'tags')
        check.str_param(mode, 'mode')

        self._environment_dict_fn = check.opt_callable_param(
            environment_dict_fn, 'environment_dict_fn'
        )
        self._should_execute = check.callable_param(should_execute, 'should_execute')

        if self._environment_dict_fn and environment_dict:
            raise DagsterInvalidDefinitionError(
                'Attempted to provide both environment_dict_fn and environment_dict as arguments'
                ' to ScheduleDefinition. Must provide only one of the two.'
            )

        self._execution_params = {
            'environmentConfigData': environment_dict,
            'selector': {'name': pipeline_name},
            'executionMetadata': {"tags": tags},
            'mode': mode,
        }

    @property
    def schedule_definition_data(self):
        return self._schedule_definition_data

    @property
    def name(self):
        return self._schedule_definition_data.name

    @property
    def cron_schedule(self):
        return self._schedule_definition_data.cron_schedule

    @property
    def environment_vars(self):
        return self._schedule_definition_data.environment_vars

    @property
    def execution_params(self):
        return self._execution_params

    @property
    def environment_dict_fn(self):
        return self._environment_dict_fn

    @property
    def should_execute(self):
        return self._should_execute



from collections import namedtuple

from dagster import check

from ..config import ConfigType
from ..field import check_field_param


class EvaluationStack(namedtuple('_EvaluationStack', 'config_type entries')):
    def __new__(cls, config_type, entries):
        return super(EvaluationStack, cls).__new__(
            cls,
            check.inst_param(config_type, 'config_type', ConfigType),
            check.list_param(entries, 'entries', of_type=EvaluationStackEntry),
        )

    @property
    def levels(self):
        return [
            entry.field_name
            for entry in self.entries
            if isinstance(entry, EvaluationStackPathEntry)
        ]

    @property
    def type_in_context(self):
        ttype = self.entries[-1].config_type if self.entries else self.config_type
        # TODO: This is the wrong place for this
        # Should have general facility for unwrapping named types
        if ttype.is_nullable:
            return ttype.inner_type
        else:
            return ttype

    def for_field(self, field_name, field_def):
        return EvaluationStack(
            config_type=self.config_type,
            entries=self.entries + [EvaluationStackPathEntry(field_name, field_def)],
        )

    def for_list_index(self, list_index):
        list_type = self.type_in_context
        check.invariant(list_type.is_list)
        return EvaluationStack(
            config_type=self.config_type,
            entries=self.entries + [EvaluationStackListItemEntry(list_type.inner_type, list_index)],
        )

    def for_set_index(self, set_index):
        set_type = self.type_in_context
        check.invariant(set_type.is_set)
        return EvaluationStack(
            config_type=self.config_type,
            entries=self.entries + [EvaluationStackListItemEntry(set_type.inner_type, set_index)],
        )

    def for_tuple_index(self, tuple_index):
        tuple_type = self.type_in_context
        check.invariant(tuple_type.is_tuple)
        return EvaluationStack(
            config_type=self.config_type,
            entries=self.entries
            + [EvaluationStackListItemEntry(tuple_type.inner_type[tuple_index], tuple_index)],
        )


class EvaluationStackEntry:  # marker interface
    pass


class EvaluationStackPathEntry(
    namedtuple('_EvaluationStackEntry', 'field_name field_def'), EvaluationStackEntry
):
    def __new__(cls, field_name, field_def):
        return super(EvaluationStackPathEntry, cls).__new__(
            cls,
            check.str_param(field_name, 'field_name'),
            check_field_param(field_def, 'field_def'),
        )

    @property
    def config_type(self):
        return self.field_def.config_type


class EvaluationStackListItemEntry(
    namedtuple('_EvaluationStackListItemEntry', 'config_type list_index'), EvaluationStackEntry
):
    def __new__(cls, config_type, list_index):
        check.int_param(list_index, 'list_index')
        check.param_invariant(list_index >= 0, 'list_index')
        return super(EvaluationStackListItemEntry, cls).__new__(
            cls, check.inst_param(config_type, 'config_type', ConfigType), list_index
        )


def get_friendly_path_msg(stack):
    return get_friendly_path_info(stack)[0]


def get_friendly_path_info(stack):
    if not stack.entries:
        path = ''
        path_msg = 'at document config root.'
    else:
        comps = ['root']
        for entry in stack.entries:
            if isinstance(entry, EvaluationStackPathEntry):
                comp = ':' + entry.field_name
                comps.append(comp)
            elif isinstance(entry, EvaluationStackListItemEntry):
                comps.append('[{i}]'.format(i=entry.list_index))
            else:
                check.failed('unsupported')

        path = ''.join(comps)
        path_msg = 'at path ' + path
    return path_msg, path



from collections import namedtuple

from dagster.utils import list_pull


def test_list_pull():
    assert list_pull([], 'foo') == []


def test_list_pull_dicts():
    test_data = [{'foo': 'bar1'}, {'foo': 'bar2'}]

    assert list_pull(test_data, 'foo') == ['bar1', 'bar2']


def test_pull_objects():
    TestObject = namedtuple('TestObject', 'bar')
    test_objs = [TestObject(bar=2), TestObject(bar=3)]
    assert list_pull(test_objs, 'bar') == [2, 3]



# encoding: utf-8
import re

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError

from .config import DEFAULT_TYPE_ATTRIBUTES, ConfigType, ConfigTypeAttributes


def all_optional_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)

    if config_type.is_composite:
        for field in config_type.fields.values():
            if not field.is_optional:
                return False
        return True

    return False


class __FieldValueSentinel:
    pass


class __InferOptionalCompositeFieldSentinel:
    pass


FIELD_NO_DEFAULT_PROVIDED = __FieldValueSentinel

INFER_OPTIONAL_COMPOSITE_FIELD = __InferOptionalCompositeFieldSentinel


def check_using_facing_field_param(obj, param_name, error_context_str):
    from dagster.core.types.field import Field, resolve_to_config_type

    check.str_param(param_name, 'param_name')
    check.str_param(error_context_str, 'error_context_str')

    if isinstance(obj, Field):
        return obj

    from .type_printer import print_config_type_to_string

    config_type = resolve_to_config_type(obj)
    if config_type:
        raise DagsterInvalidDefinitionError(
            (
                'You have passed a config type "{printed_type}" in the parameter '
                '"{param_name}" {error_context_str}. '
                'You have likely forgot to wrap this type in a Field.'
            ).format(
                printed_type=print_config_type_to_string(config_type, with_lines=False),
                error_context_str=error_context_str,
                param_name=param_name,
            )
        )
    else:
        raise DagsterInvalidDefinitionError(
            (
                'You have passed an object {value_repr} of incorrect type '
                '"{type_name}" in the parameter "{param_name}" '
                '{error_context_str} where a Field was expected.'
            ).format(
                error_context_str=error_context_str,
                param_name=param_name,
                value_repr=repr(obj),
                type_name=type(obj).__name__,
            )
        )


def check_user_facing_opt_field_param(obj, param_name, error_context_str):
    check.str_param(param_name, 'param_name')
    check.str_param(error_context_str, 'error_context_str')
    if obj is None:
        return None
    return check_using_facing_field_param(obj, param_name, error_context_str)


class _ConfigHasFields(ConfigType):
    def __init__(self, fields, *args, **kwargs):
        from dagster.core.types.field import Field

        self.fields = check.dict_param(fields, 'fields', key_type=str, value_type=Field)
        super(_ConfigHasFields, self).__init__(*args, **kwargs)

    @property
    def inner_types(self):
        return list(set(self._yield_inner_types()))

    def _yield_inner_types(self):
        for field in self.fields.values():
            yield field.config_type
            for inner_type in field.config_type.inner_types:
                yield inner_type

    def debug_str(self):
        def left_pad(line, n=2, char=" "):
            lines = line.splitlines()
            return "\n".join(map(lambda x: (char * n) + x, lines))

        def format_fields(fields):
            lines = map(left_pad, fields)
            return '{{\n{lines}\n}}'.format(lines="\n".join(lines))

        def field_to_string(field_key, field_value):
            s = "{k}{opt}: ".format(k=field_key, opt="?" if field_value.is_optional else "")

            if field_value.config_type.is_scalar:
                typ = re.search(r'\.([^\.]*) object', str(field_value.config_type)).group(1)
                s += "[" + typ + "] "
            elif field_value.config_type.is_selector:
                fields = [
                    field_to_string("-" + k, v)
                    for k, v in field_value.config_type.inst().fields.items()
                ]
                fields = map(lambda x: left_pad(x, 1, "|"), fields)
                s += format_fields(fields)
            elif field_value.config_type.is_enum:
                fields = [
                    field_to_string(k, v) for k, v in field_value.config_type.inst().fields.items()
                ]
                s += format_fields(fields)
            elif field_value.config_type.is_composite:
                fields = [
                    field_to_string(k, v) for k, v in field_value.config_type.inst().fields.items()
                ]
                s += format_fields(fields)
            elif field_value.config_type.is_list:
                fields = [
                    field_to_string(k, v)
                    for k, v in field_value.config_type.inst().inner_type.inst().fields.items()
                ]
                s += format_fields(fields)
            elif field_value.config_type.is_any:
                s += "[Any]"

            s += (
                " default=" + str(field_value.default_value) if field_value.default_provided else ""
            )
            return s

        fields = [field_to_string(k, v) for k, v in self.fields.items()]
        return format_fields(fields)


class _ConfigComposite(_ConfigHasFields):
    @property
    def is_composite(self):
        return True

    @property
    def is_permissive_composite(self):
        return False


class _CompositeSolidConfigComposite(_ConfigComposite):
    @property
    def handle(self):
        return self.type_attributes.handle


class _ConfigSelector(_ConfigHasFields):
    @property
    def is_selector(self):
        return True


# HACK HACK HACK
#
# This is not good and a better solution needs to be found. In order
# for the client-side typeahead in dagit to work as currently structured,
# dictionaries need names. While we deal with that we're going to automatically
# name dictionaries. This will cause odd behavior and bugs is you restart
# the server-side process, the type names changes, and you do not refresh the client.
#
# A possible short term mitigation would to name the dictionary based on the hash
# of its member fields to provide stability in between process restarts.
#
class DictCounter:
    _count = 0

    @staticmethod
    def get_next_count():
        DictCounter._count += 1
        return DictCounter._count


def check_user_facing_fields_dict(fields, type_name_msg):
    check.dict_param(fields, 'fields', key_type=str)
    check.str_param(type_name_msg, 'type_name_msg')

    sorted_field_names = sorted(list(fields.keys()))
    for field_name, potential_field in fields.items():
        check_using_facing_field_param(
            potential_field,
            'fields',
            (
                'and it is in the "{field_name}" entry of that dict. It is from '
                'a {type_name_msg} with fields {field_names}'
            ).format(
                type_name_msg=type_name_msg, field_name=field_name, field_names=sorted_field_names
            ),
        )


def NamedDict(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    '''
    A :py:class:`Dict` with a name allowing it to be referenced by that name.
    '''
    check_user_facing_fields_dict(fields, 'NamedDict named "{}"'.format(name))

    class _NamedDict(_ConfigComposite):
        def __init__(self):
            super(_NamedDict, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedDict


class DictTypeApi:
    def __call__(self, fields):
        return build_config_dict(fields)

    def __getitem__(self, *args):
        from .python_dict import create_typed_runtime_dict

        check.param_invariant(len(args[0]) == 2, 'args', 'Must be two parameters')
        return create_typed_runtime_dict(args[0][0], args[0][1])


Dict = DictTypeApi()


def build_config_dict(fields):
    '''
    Schema for configuration data with string keys and typed values via :py:class:`Field` .

    Args:
        fields (Dict[str, Field])
    '''
    check_user_facing_fields_dict(fields, 'Dict')

    class _Dict(_ConfigComposite):
        def __init__(self):
            key = 'Dict.' + str(DictCounter.get_next_count())
            super(_Dict, self).__init__(
                name=None,
                key=key,
                fields=fields,
                description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Dict


def PermissiveDict(fields=None):
    '''Defines a config dict with a partially specified schema.
    
    A permissive dict allows partial specification of the config schema. Any fields with a
    specified schema will be type checked. Other fields will be allowed, but will be ignored by
    the type checker.

    Args:
        fields (Dict[str, Field]): The partial specification of the config dict.
    
    **Examples**

    .. code-block:: python

        @solid(config_field=Field(PermissiveDict({'required': Field(String)})))
        def partially_specified_config(context) -> List:
            return sorted(list(context.solid_config.items()))
    '''

    if fields:
        check_user_facing_fields_dict(fields, 'PermissiveDict')

    class _PermissiveDict(_ConfigComposite):
        def __init__(self):
            key = 'PermissiveDict.' + str(DictCounter.get_next_count())
            super(_PermissiveDict, self).__init__(
                name=None,
                key=key,
                fields=fields or dict(),
                description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

        @property
        def is_permissive_composite(self):
            return True

    return _PermissiveDict


def Selector(fields):
    '''Define a config field requiring the user to select one option.
    
    Selectors are used when you want to be able to present several different options in config but
    allow only one to be selected. For example, a single input might be read in from either a csv
    file or a parquet file, but not both at once.

    Note that in some other type systems this might be called an 'input union'.

    Functionally, a selector is like a :py:class:`Dict`, except that only one key from the dict can
    be specified in valid config.

    Args:
        fields (Dict[str, Field]): The fields from which the user must select.
    
    **Examples**:

    .. code-block:: python

        @solid(
            config_field=Field(
                Selector(
                    {
                        'haw': Field(
                            Dict({'whom': Field(String, default_value='honua', is_optional=True)})
                        ),
                        'cn': Field(
                            Dict({'whom': Field(String, default_value='世界', is_optional=True)})
                        ),
                        'en': Field(
                            Dict({'whom': Field(String, default_value='world', is_optional=True)})
                        )
                    }
                ),
                is_optional=True,
                default_value={'en': {'whom': 'world'}},
            )
        )
        def hello_world_with_default(context):
            if 'haw' in context.solid_config:
                return 'Aloha {whom}!'.format(whom=context.solid_config['haw']['whom'])
            if 'cn' in context.solid_config:
                return '你好，{whom}!'.format(whom=context.solid_config['cn']['whom'])
            if 'en' in context.solid_config:
                return 'Hello, {whom}!'.format(whom=context.solid_config['en']['whom'])
    '''
    check_user_facing_fields_dict(fields, 'Selector')

    class _Selector(_ConfigSelector):
        def __init__(self):
            key = 'Selector.' + str(DictCounter.get_next_count())
            super(_Selector, self).__init__(
                key=key,
                name=None,
                fields=fields,
                # description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Selector


def NamedSelector(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    '''
    A :py:class`Selector` with a name, allowing it to be referenced by that name.

    Args:
        name (str):
        fields (Dict[str, Field])
    '''
    check.str_param(name, 'name')
    check_user_facing_fields_dict(fields, 'NamedSelector named "{}"'.format(name))

    class _NamedSelector(_ConfigSelector):
        def __init__(self):
            super(_NamedSelector, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedSelector



import pytest

from dagster import (
    Any,
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    Int,
    ModeDefinition,
    OutputDefinition,
    PipelineDefinition,
    ResourceDefinition,
    SolidDefinition,
    SolidInvocation,
    String,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.core.definitions import create_environment_schema, create_environment_type
from dagster.core.definitions.environment_configs import (
    EnvironmentClassCreationData,
    define_resource_cls,
    define_solid_config_cls,
    define_solid_dictionary_cls,
)
from dagster.core.system_config.objects import (
    EnvironmentConfig,
    SolidConfig,
    construct_solid_dictionary,
)
from dagster.core.test_utils import throwing_evaluate_config_value
from dagster.core.types.evaluator import evaluate_config
from dagster.core.types.evaluator.errors import DagsterEvaluateConfigValueError
from dagster.core.types.field_utils import NamedDict
from dagster.loggers import default_loggers


def create_creation_data(pipeline_def):
    return EnvironmentClassCreationData(
        pipeline_def.name,
        pipeline_def.solids,
        pipeline_def.dependency_structure,
        mode_definition=None,
        logger_defs=default_loggers(),
    )


def test_resource_config_any():
    resource_type = define_resource_cls(
        'Parent', 'Foo', ResourceDefinition(lambda: None, Field(Any))
    ).inst()

    assert resource_type.type_attributes.is_system_config

    assert throwing_evaluate_config_value(resource_type, {'config': 1}) == {'config': 1}


def test_all_types_provided():
    pipeline_def = PipelineDefinition(
        name='pipeline',
        solid_defs=[],
        mode_defs=[
            ModeDefinition(
                name='SomeMode',
                resource_defs={
                    'some_resource': ResourceDefinition(
                        lambda: None,
                        config_field=Field(
                            NamedDict(
                                'SomeModeNamedDict',
                                {
                                    'with_default_int': Field(
                                        Int, is_optional=True, default_value=23434
                                    )
                                },
                            )
                        ),
                    )
                },
            )
        ],
    )

    environment_schema = create_environment_schema(pipeline_def)

    all_types = list(environment_schema.all_config_types())
    type_names = set(t.name for t in all_types)
    assert 'SomeModeNamedDict' in type_names
    assert 'Pipeline.Mode.SomeMode.Environment' in type_names
    assert 'Pipeline.Mode.SomeMode.Resources.SomeResource' in type_names


def test_provided_default_on_resources_config():
    @solid(name='some_solid', input_defs=[], output_defs=[])
    def some_solid(_):
        return None

    @pipeline(
        mode_defs=[
            ModeDefinition(
                name='some_mode',
                resource_defs={
                    'some_resource': ResourceDefinition(
                        resource_fn=lambda: None,
                        config_field=Field(
                            Dict(
                                {
                                    'with_default_int': Field(
                                        Int, is_optional=True, default_value=23434
                                    )
                                }
                            )
                        ),
                    )
                },
            )
        ]
    )
    def pipeline_def():
        some_solid()

    env_type = create_environment_type(pipeline_def)
    assert env_type.type_attributes.is_system_config
    some_resource_field = env_type.fields['resources'].config_type.fields['some_resource']
    assert some_resource_field.is_optional

    some_resource_config_field = some_resource_field.config_type.fields['config']
    assert some_resource_config_field.is_optional
    assert some_resource_config_field.default_value == {'with_default_int': 23434}

    assert some_resource_field.default_value == {'config': {'with_default_int': 23434}}

    value = EnvironmentConfig.from_config_value(throwing_evaluate_config_value(env_type, {}), {})
    assert value.resources == {'some_resource': {'config': {'with_default_int': 23434}}}


def test_default_environment():
    @solid(name='some_solid', input_defs=[], output_defs=[])
    def some_solid(_):
        return None

    @pipeline
    def pipeline_def():
        some_solid()

    EnvironmentConfig.from_config_value(
        throwing_evaluate_config_value(create_environment_type(pipeline_def), {}), {}
    )


def test_resource_def_config_errors():
    takes_int_resource_def = ResourceDefinition(
        resource_fn=lambda: None, config_field=Field(Dict({'required_int': Field(Int)}))
    )

    resource_type = define_resource_cls('Parent', 'takes_int', takes_int_resource_def).inst()

    assert not evaluate_config(resource_type, 1).success
    assert not evaluate_config(resource_type, {}).success
    assert not evaluate_config(resource_type, {'config': {}}).success
    assert evaluate_config(resource_type, {'config': {'required_int': 2}}).success
    assert not evaluate_config(resource_type, {'config': {'required_int': 'kdjfkd'}}).success


def test_solid_config():
    solid_config_type = define_solid_config_cls('kdjfkd', Field(Int), None, None).inst()
    solid_inst = throwing_evaluate_config_value(solid_config_type, {'config': 1})
    assert solid_inst['config'] == 1
    assert solid_config_type.inst().type_attributes.is_system_config


def test_solid_dictionary_type():
    pipeline_def = define_test_solids_config_pipeline()

    solid_dict_type = define_solid_dictionary_cls(
        'foobar', pipeline_def.solids, pipeline_def.dependency_structure, pipeline_def.name
    ).inst()

    value = construct_solid_dictionary(
        throwing_evaluate_config_value(
            solid_dict_type,
            {'int_config_solid': {'config': 1}, 'string_config_solid': {'config': 'bar'}},
        )
    )

    assert set(['int_config_solid', 'string_config_solid']) == set(value.keys())
    assert value == {'int_config_solid': SolidConfig(1), 'string_config_solid': SolidConfig('bar')}

    assert solid_dict_type.type_attributes.is_system_config

    for specific_solid_config_field in solid_dict_type.fields.values():
        specific_solid_config_type = specific_solid_config_field.config_type
        assert specific_solid_config_type.type_attributes.is_system_config
        user_config_field = specific_solid_config_field.config_type.fields['config']
        assert user_config_field.config_type.type_attributes.is_system_config is False


def define_test_solids_config_pipeline():
    @solid(
        name='int_config_solid',
        config_field=Field(Int, is_optional=True),
        input_defs=[],
        output_defs=[],
    )
    def int_config_solid(_):
        return None

    @solid(
        name='string_config_solid',
        config_field=Field(String, is_optional=True),
        input_defs=[],
        output_defs=[],
    )
    def string_config_solid(_):
        return None

    @pipeline
    def pipeline_def():
        int_config_solid()
        string_config_solid()

    return pipeline_def


def assert_has_fields(dtype, *fields):
    return set(dtype.fields.keys()) == set(fields)


def test_solid_configs_defaults():
    env_type = create_environment_type(define_test_solids_config_pipeline())

    solids_field = env_type.fields['solids']

    assert_has_fields(solids_field.config_type, 'int_config_solid', 'string_config_solid')

    int_solid_field = solids_field.config_type.fields['int_config_solid']

    assert int_solid_field.is_optional
    # TODO: this is the test case the exposes the default dodginess
    assert int_solid_field.default_provided

    assert_has_fields(int_solid_field.config_type, 'config')

    int_solid_config_field = int_solid_field.config_type.fields['config']

    assert int_solid_config_field.is_optional
    assert not int_solid_config_field.default_provided


def test_solid_dictionary_some_no_config():
    @solid(name='int_config_solid', config_field=Field(Int), input_defs=[], output_defs=[])
    def int_config_solid(_):
        return None

    @solid(name='no_config_solid', input_defs=[], output_defs=[])
    def no_config_solid(_):
        return None

    @pipeline
    def pipeline_def():
        int_config_solid()
        no_config_solid()

    solid_dict_type = define_solid_dictionary_cls(
        'foobar', pipeline_def.solids, pipeline_def.dependency_structure, pipeline_def.name
    ).inst()

    value = construct_solid_dictionary(
        throwing_evaluate_config_value(solid_dict_type, {'int_config_solid': {'config': 1}})
    )

    assert set(['int_config_solid']) == set(value.keys())
    assert value == {'int_config_solid': SolidConfig(1)}


def test_whole_environment():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        mode_defs=[
            ModeDefinition(
                name='test_mode',
                resource_defs={
                    'test_resource': ResourceDefinition(
                        resource_fn=lambda: None, config_field=Field(Any)
                    )
                },
            )
        ],
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda *args: None,
            ),
            SolidDefinition(
                name='no_config_solid', input_defs=[], output_defs=[], compute_fn=lambda *args: None
            ),
        ],
    )

    environment_type = create_environment_type(pipeline_def)

    assert (
        environment_type.fields['resources'].config_type.name
        == 'SomePipeline.Mode.TestMode.Resources'
    )
    solids_type = environment_type.fields['solids'].config_type
    assert solids_type.name == 'SomePipeline.SolidsConfigDictionary'
    assert (
        solids_type.fields['int_config_solid'].config_type.name
        == 'SomePipeline.SolidConfig.IntConfigSolid'
    )

    env = EnvironmentConfig.from_config_value(
        throwing_evaluate_config_value(
            environment_type,
            {
                'resources': {'test_resource': {'config': 1}},
                'solids': {'int_config_solid': {'config': 123}},
            },
        ),
        {
            'resources': {'test_resource': {'config': 1}},
            'solids': {'int_config_solid': {'config': 123}},
        },
    )

    assert isinstance(env, EnvironmentConfig)
    assert env.solids == {'int_config_solid': SolidConfig(123)}
    assert env.resources == {'test_resource': {'config': 1}}


def test_solid_config_error():
    pipeline_def = define_test_solids_config_pipeline()
    solid_dict_type = define_solid_dictionary_cls(
        'slkdfjkjdsf', pipeline_def.solids, pipeline_def.dependency_structure, pipeline_def.name
    ).inst()

    int_solid_config_type = solid_dict_type.fields['int_config_solid'].config_type

    with pytest.raises(DagsterEvaluateConfigValueError, match='Field "notconfig" is not defined'):
        throwing_evaluate_config_value(int_solid_config_type, {'notconfig': 1})

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(int_solid_config_type, 1)


def test_optional_solid_with_no_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda context, _inputs: _assert_config_none(context, 234),
            ),
            SolidDefinition(
                name='no_config_solid',
                input_defs=[],
                output_defs=[],
                compute_fn=lambda context, _inputs: _assert_config_none(context, None),
            ),
        ],
    )

    assert execute_pipeline(pipeline_def, {'solids': {'int_config_solid': {'config': 234}}}).success


def test_optional_solid_with_optional_scalar_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int, is_optional=True),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda context, _inputs: _assert_config_none(context, 234),
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)

    assert env_type.fields['solids'].is_optional is True

    solids_type = env_type.fields['solids'].config_type

    assert solids_type.fields['int_config_solid'].is_optional is True

    solids_default_obj = construct_solid_dictionary(throwing_evaluate_config_value(solids_type, {}))

    assert solids_default_obj['int_config_solid'].config is None

    env_obj = EnvironmentConfig.from_config_value(throwing_evaluate_config_value(env_type, {}), {})

    assert env_obj.solids['int_config_solid'].config is None


def test_optional_solid_with_required_scalar_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda context, _inputs: _assert_config_none(context, 234),
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)

    assert env_type.fields['solids'].is_optional is False

    solids_type = env_type.fields['solids'].config_type

    assert solids_type.fields['int_config_solid'].is_optional is False

    int_config_solid_type = solids_type.fields['int_config_solid'].config_type

    assert_has_fields(int_config_solid_type, 'config')

    int_config_solid_config_field = int_config_solid_type.fields['config']

    assert int_config_solid_config_field.is_optional is False

    execute_pipeline(pipeline_def, {'solids': {'int_config_solid': {'config': 234}}})


def test_required_solid_with_required_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Dict({'required_field': Field(String)})),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda *_args: None,
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)

    assert env_type.fields['solids'].is_optional is False
    assert env_type.fields['solids'].config_type

    solids_type = env_type.fields['solids'].config_type
    assert solids_type.fields['int_config_solid'].is_optional is False
    int_config_solid_type = solids_type.fields['int_config_solid'].config_type
    assert int_config_solid_type.fields['config'].is_optional is False

    assert env_type.fields['execution'].is_optional

    env_obj = EnvironmentConfig.from_config_value(
        throwing_evaluate_config_value(
            env_type, {'solids': {'int_config_solid': {'config': {'required_field': 'foobar'}}}}
        ),
        {'solids': {'int_config_solid': {'config': {'required_field': 'foobar'}}}},
    )

    assert env_obj.solids['int_config_solid'].config['required_field'] == 'foobar'

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(env_type, {'solids': {}})

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(env_type, {})


def test_optional_solid_with_optional_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(
                    Dict({'optional_field': Field(String, is_optional=True)}), is_optional=True
                ),
                input_defs=[],
                output_defs=[],
                compute_fn=lambda *_args: None,
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['execution'].is_optional


def nested_field(config_type, *field_names):
    assert field_names

    field = config_type.fields[field_names[0]]

    for field_name in field_names[1:]:
        field = field.config_type.fields[field_name]

    return field


def test_required_resource_with_required_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'with_required': ResourceDefinition(
                        resource_fn=lambda: None,
                        config_field=Field(Dict({'required_field': Field(String)})),
                    )
                }
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['resources'].is_required
    assert nested_field(env_type, 'resources', 'with_required').is_required
    assert nested_field(env_type, 'resources', 'with_required', 'config').is_required
    assert nested_field(
        env_type, 'resources', 'with_required', 'config', 'required_field'
    ).is_required


def test_all_optional_field_on_single_resource():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'with_optional': ResourceDefinition(
                        resource_fn=lambda: None,
                        config_field=Field(
                            Dict({'optional_field': Field(String, is_optional=True)})
                        ),
                    )
                }
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['resources'].is_optional
    assert nested_field(env_type, 'resources', 'with_optional').is_optional
    assert nested_field(env_type, 'resources', 'with_optional', 'config').is_optional
    assert nested_field(
        env_type, 'resources', 'with_optional', 'config', 'optional_field'
    ).is_optional


def test_optional_and_required_context():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solid_defs=[],
        mode_defs=[
            ModeDefinition(
                name='mixed',
                resource_defs={
                    'optional_resource': ResourceDefinition(
                        lambda: None,
                        config_field=Field(
                            dagster_type=Dict(
                                fields={'optional_field': Field(String, is_optional=True)}
                            )
                        ),
                    ),
                    'required_resource': ResourceDefinition(
                        lambda: None,
                        config_field=Field(
                            dagster_type=Dict(fields={'required_field': Field(String)})
                        ),
                    ),
                },
            )
        ],
    )

    env_type = create_environment_type(pipeline_def)
    assert env_type.fields['solids'].is_optional

    assert env_type.fields['execution'].is_optional

    assert nested_field(env_type, 'resources').is_required
    assert nested_field(env_type, 'resources', 'optional_resource').is_optional
    assert nested_field(env_type, 'resources', 'optional_resource', 'config').is_optional
    assert nested_field(
        env_type, 'resources', 'optional_resource', 'config', 'optional_field'
    ).is_optional

    assert nested_field(env_type, 'resources', 'required_resource').is_required
    assert nested_field(env_type, 'resources', 'required_resource', 'config').is_required
    assert nested_field(
        env_type, 'resources', 'required_resource', 'config', 'required_field'
    ).is_required

    env_obj = EnvironmentConfig.from_config_value(
        throwing_evaluate_config_value(
            env_type, {'resources': {'required_resource': {'config': {'required_field': 'foo'}}}}
        ),
        {'resources': {'required_resource': {'config': {'required_field': 'foo'}}}},
    )

    assert env_obj.resources == {
        'optional_resource': {'config': {}},
        'required_resource': {'config': {'required_field': 'foo'}},
    }


def test_required_inputs():
    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    pipeline_def = PipelineDefinition(
        name='required_int_input',
        solid_defs=[add_one],
        dependencies={
            SolidInvocation('add_one', 'first_add'): {},
            SolidInvocation('add_one', 'second_add'): {'num': DependencyDefinition('first_add')},
        },
    )

    env_type = create_environment_type(pipeline_def)

    solids_type = env_type.fields['solids'].config_type

    first_add_fields = solids_type.fields['first_add'].config_type.fields

    assert 'inputs' in first_add_fields

    inputs_field = first_add_fields['inputs']

    assert inputs_field.is_required

    assert inputs_field.config_type.fields['num'].is_required

    # second_add has a dependency so the input is not available
    assert 'inputs' not in solids_type.fields['second_add'].config_type.fields


def test_mix_required_inputs():
    @lambda_solid(
        input_defs=[InputDefinition('left', Int), InputDefinition('right', Int)],
        output_def=OutputDefinition(Int),
    )
    def add_numbers(left, right):
        return left + right

    @lambda_solid
    def return_three():
        return 3

    pipeline_def = PipelineDefinition(
        name='mixed_required_inputs',
        solid_defs=[add_numbers, return_three],
        dependencies={'add_numbers': {'right': DependencyDefinition('return_three')}},
    )

    env_type = create_environment_type(pipeline_def)
    solids_type = env_type.fields['solids'].config_type
    add_numbers_type = solids_type.fields['add_numbers'].config_type
    inputs_fields_dict = add_numbers_type.fields['inputs'].config_type.fields

    assert 'left' in inputs_fields_dict
    assert not 'right' in inputs_fields_dict


def test_files_default_config():
    pipeline_def = PipelineDefinition(name='pipeline', solid_defs=[])

    env_type = create_environment_type(pipeline_def)
    assert 'storage' in env_type.fields

    config_value = throwing_evaluate_config_value(env_type, {})

    assert 'storage' not in config_value


def test_storage_in_memory_config():
    pipeline_def = PipelineDefinition(name='pipeline', solid_defs=[])

    env_type = create_environment_type(pipeline_def)
    assert 'storage' in env_type.fields

    config_value = throwing_evaluate_config_value(env_type, {'storage': {'in_memory': {}}})

    assert config_value['storage'] == {'in_memory': {}}


def test_directly_init_environment_config():
    EnvironmentConfig()



import uuid

import toposort as toposort_


def toposort(data):
    return [sorted(list(level)) for level in toposort_.toposort(data)]


def toposort_flatten(data):
    return [item for level in toposort(data) for item in level]


def make_new_run_id():
    return str(uuid.uuid4())



import pytest

from dagster import DagsterInvalidConfigError, Field, String, execute_pipeline, pipeline, solid


def test_basic_solid_with_config():
    did_get = {}

    @solid(
        name='solid_with_context',
        input_defs=[],
        output_defs=[],
        config={'some_config': Field(String)},
    )
    def solid_with_context(context):
        did_get['yep'] = context.solid_config

    @pipeline
    def pipeline_def():
        solid_with_context()

    execute_pipeline(
        pipeline_def, {'solids': {'solid_with_context': {'config': {'some_config': 'foo'}}}}
    )

    assert 'yep' in did_get
    assert 'some_config' in did_get['yep']


def test_config_arg_mismatch():
    def _t_fn(*_args):
        raise Exception('should not reach')

    @solid(
        name='solid_with_context',
        input_defs=[],
        output_defs=[],
        config={'some_config': Field(String)},
    )
    def solid_with_context(context):
        raise Exception('should not reach')

    @pipeline
    def pipeline_def():
        solid_with_context()

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(
            pipeline_def, {'solids': {'solid_with_context': {'config': {'some_config': 1}}}}
        )


def test_solid_not_found():
    @solid(name='find_me_solid', input_defs=[], output_defs=[])
    def find_me_solid(_):
        raise Exception('should not reach')

    @pipeline
    def pipeline_def():
        find_me_solid()

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(pipeline_def, {'solids': {'not_found': {'config': {'some_config': 1}}}})


def test_config_for_no_config():
    @solid(name='no_config_solid', input_defs=[], output_defs=[])
    def no_config_solid(_):
        raise Exception('should not reach')

    @pipeline
    def pipeline_def():
        return no_config_solid()

    with pytest.raises(DagsterInvalidConfigError):
        execute_pipeline(
            pipeline_def, {'solids': {'no_config_solid': {'config': {'some_config': 1}}}}
        )



from __future__ import print_function

import sys

import click
import six

from dagster import check
from dagster.core.execution.compute_logs import (
    should_compute_log_tail_poll,
    tail_polling,
    tail_posix,
)


def create_utils_cli_group():
    group = click.Group(name="utils")
    group.add_command(tail_command)
    return group


def extract_string_param(name, value):
    if value and not isinstance(value, six.string_types):
        if len(value) == 1:
            return value[0]
        else:
            check.failed(
                'Can only handle zero or one {name} args. Got {value}'.format(
                    name=repr(name), value=repr(value)
                )
            )
    return value


@click.command(name='tail', help='Cross-platform file tailing')
@click.argument('filepath')
@click.option('--parent-pid', help="Process pid to watch for termination")
@click.option('--io-type', help="Whether to stream to stdout/stderr", default='stdout')
def tail_command(filepath, parent_pid, io_type):
    filepath = extract_string_param('filepath', filepath)
    parent_pid = extract_string_param('parent_pid', parent_pid)
    io_type = extract_string_param('io_type', io_type)
    stream = sys.stdout if io_type == 'stdout' else sys.stderr

    if should_compute_log_tail_poll():
        tail_polling(filepath, stream, parent_pid)
    else:
        tail_posix(filepath, stream, parent_pid)


utils_cli = create_utils_cli_group()



import io
import os
import shutil
from abc import ABCMeta, abstractmethod

import six

from dagster import check
from dagster.core.definitions import resource
from dagster.core.types import Field
from dagster.utils import mkdir_p

from .file_manager import LocalFileHandle


class FileCache(six.with_metaclass(ABCMeta)):
    def __init__(self, overwrite):
        # Overwrite is currently only a signal to callers to not overwrite.
        # These classes currently do not enforce any semantics around that
        self.overwrite = check.bool_param(overwrite, 'overwrite')

    @abstractmethod
    def has_file_object(self, file_key):
        pass

    @abstractmethod
    def get_file_handle(self, file_key):
        pass

    @abstractmethod
    def write_file_object(self, file_key, source_file_object):
        pass

    def write_binary_data(self, file_key, binary_data):
        return self.write_file_object(file_key, io.BytesIO(binary_data))


class FSFileCache(FileCache):
    def __init__(self, target_folder, overwrite=False):
        super(FSFileCache, self).__init__(overwrite=overwrite)
        check.str_param(target_folder, 'target_folder')
        check.param_invariant(os.path.isdir(target_folder), 'target_folder')

        self.target_folder = target_folder

    def has_file_object(self, file_key):
        return os.path.exists(self.get_full_path(file_key))

    def write_file_object(self, file_key, source_file_object):
        target_file = self.get_full_path(file_key)
        with open(target_file, 'wb') as dest_file_object:
            shutil.copyfileobj(source_file_object, dest_file_object)
        return LocalFileHandle(target_file)

    def get_file_handle(self, file_key):
        check.str_param(file_key, 'file_key')
        return LocalFileHandle(self.get_full_path(file_key))

    def get_full_path(self, file_key):
        check.str_param(file_key, 'file_key')
        return os.path.join(self.target_folder, file_key)


@resource(
    {'overwrite': Field(bool, is_optional=True, default_value=False), 'target_folder': Field(str)}
)
def fs_file_cache(init_context):
    target_folder = init_context.resource_config['target_folder']

    if not os.path.exists(target_folder):
        mkdir_p(target_folder)

    return FSFileCache(target_folder=target_folder, overwrite=False)



from dagster.core.definitions import (
    CompositeSolidDefinition,
    ConfigMapping,
    ConfigMappingContext,
    DependencyDefinition,
    EventMetadataEntry,
    ExecutionTargetHandle,
    ExecutorDefinition,
    ExpectationResult,
    Failure,
    InputDefinition,
    InputMapping,
    JsonMetadataEntryData,
    LoggerDefinition,
    MarkdownMetadataEntryData,
    Materialization,
    ModeDefinition,
    MultiDependencyDefinition,
    Output,
    OutputDefinition,
    OutputMapping,
    PathMetadataEntryData,
    PipelineDefinition,
    PresetDefinition,
    RepositoryDefinition,
    ResourceDefinition,
    ScheduleDefinition,
    SolidDefinition,
    SolidInvocation,
    SystemStorageData,
    SystemStorageDefinition,
    TextMetadataEntryData,
    TypeCheck,
    UrlMetadataEntryData,
    composite_solid,
    executor,
    lambda_solid,
    logger,
    pipeline,
    resource,
    schedules,
    solid,
    system_storage,
)
from dagster.core.engine import Engine
from dagster.core.engine.init import InitExecutorContext
from dagster.core.errors import (
    DagsterError,
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterResourceFunctionError,
    DagsterRunNotFoundError,
    DagsterStepOutputNotFoundError,
    DagsterSubprocessError,
    DagsterTypeCheckError,
    DagsterUnmetExecutorRequirementsError,
    DagsterUserCodeExecutionError,
)
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.api import (
    execute_pipeline,
    execute_pipeline_iterator,
    execute_pipeline_with_preset,
)
from dagster.core.execution.config import ExecutorConfig, RunConfig
from dagster.core.execution.context.init import InitResourceContext
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.core.execution.context.system import SystemComputeExecutionContext
from dagster.core.execution.plan.objects import StepKind
from dagster.core.execution.results import PipelineExecutionResult, SolidExecutionResult
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.file_manager import FileHandle, LocalFileHandle
from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.types import (
    Any,
    Bool,
    Dict,
    Field,
    Float,
    Int,
    List,
    Nothing,
    Optional,
    Path,
    PermissiveDict,
    Selector,
    Set,
    String,
    Tuple,
    input_hydration_config,
    output_materialization_config,
)
from dagster.core.types.config import Enum, EnumValue
from dagster.core.types.decorator import as_dagster_type, dagster_type
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.types.runtime import define_python_dagster_type
from dagster.utils import file_relative_path
from dagster.utils.test import (
    check_dagster_type,
    execute_solid,
    execute_solid_within_pipeline,
    execute_solids_within_pipeline,
)

from .version import __version__

__all__ = [
    # Definition
    'ConfigMapping',
    'ConfigMappingContext',
    'CompositeSolidDefinition',
    'DependencyDefinition',
    'EventMetadataEntry',
    'ExecutorDefinition',
    'TextMetadataEntryData',
    'UrlMetadataEntryData',
    'PathMetadataEntryData',
    'JsonMetadataEntryData',
    'ExpectationResult',
    'Failure',
    'Field',
    'InputDefinition',
    'InputMapping',
    'LoggerDefinition',
    'Materialization',
    'ModeDefinition',
    'OutputDefinition',
    'OutputMapping',
    'PipelineDefinition',
    'RepositoryDefinition',
    'ResourceDefinition',
    'Output',
    'SolidDefinition',
    'SolidInvocation',
    'SystemStorageDefinition',
    # Decorators
    'composite_solid',
    'executor',
    'lambda_solid',
    'logger',
    'pipeline',
    'resource',
    'solid',
    'system_storage',
    # Execution
    'execute_pipeline_iterator',
    'execute_pipeline',
    'DagsterEvent',
    'DagsterEventType',
    'Engine',
    'ExecutorConfig',
    'InitExecutorContext',
    'InitLoggerContext',
    'InitResourceContext',
    'InitSystemStorageContext',
    'PipelineExecutionResult',
    'RunConfig',
    'SolidExecutionResult',
    'SystemComputeExecutionContext',
    'SystemStorageData',
    # Errors
    'DagsterExecutionStepExecutionError',
    'DagsterExecutionStepNotFoundError',
    'DagsterInvalidConfigError',
    'DagsterInvalidDefinitionError',
    'DagsterInvariantViolationError',
    'DagsterResourceFunctionError',
    'DagsterTypeCheckError',
    'DagsterUserCodeExecutionError',
    'DagsterStepOutputNotFoundError',
    'DagsterSubprocessError',
    # Logging
    'DagsterLogManager',
    # Utilities
    'check_dagster_type',
    'execute_solid',
    'execute_solids_within_pipeline',
    'file_relative_path',
    # types
    'Any',
    'Bool',
    'input_hydration_config',
    'Dict',
    'Enum',
    'EnumValue',
    'Float',
    'Int',
    'List',
    'Optional',
    'output_materialization_config',
    'Path',
    'PermissiveDict',
    'String',
    'SerializationStrategy',
    'Nothing',
    # type creation
    'as_dagster_type',
    'dagster_type',
    'define_python_dagster_type',
    # file things
    'FileHandle',
    'LocalFileHandle',
]



from abc import abstractmethod
from collections import OrderedDict, defaultdict
from datetime import datetime

import sqlalchemy as db

from dagster import check
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple

from .pipeline_run import PipelineRun, PipelineRunStatus
from .run_storage_abc import RunStorage


class InMemoryRunStorage(RunStorage):
    def __init__(self):
        self._runs = OrderedDict()

    def add_run(self, pipeline_run):
        check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
        check.invariant(
            not self._runs.get(pipeline_run.run_id),
            'Can not add same run twice for run_id {run_id}'.format(run_id=pipeline_run.run_id),
        )

        self._runs[pipeline_run.run_id] = pipeline_run
        return pipeline_run

    def handle_run_event(self, run_id, event):
        check.str_param(run_id, 'run_id')
        check.inst_param(event, 'event', DagsterEvent)
        run = self._runs[run_id]

        if event.event_type == DagsterEventType.PIPELINE_START:
            self._runs[run_id] = run.run_with_status(PipelineRunStatus.STARTED)
        elif event.event_type == DagsterEventType.PIPELINE_SUCCESS:
            self._runs[run_id] = run.run_with_status(PipelineRunStatus.SUCCESS)
        elif event.event_type == DagsterEventType.PIPELINE_FAILURE:
            self._runs[run_id] = self._runs[run_id].run_with_status(PipelineRunStatus.FAILURE)

    def all_runs(self, cursor=None, limit=None):
        return self._slice(list(self._runs.values())[::-1], cursor, limit)

    def get_runs_with_pipeline_name(self, pipeline_name, cursor=None, limit=None):
        check.str_param(pipeline_name, 'pipeline_name')
        return self._slice(
            [r for r in self.all_runs() if r.pipeline_name == pipeline_name], cursor, limit
        )

    def get_run_count_with_matching_tags(self, tags):
        check.list_param(tags, 'tags', tuple)
        return len(self.get_runs_with_matching_tags(tags))

    def get_runs_with_matching_tags(self, tags, cursor=None, limit=None):
        check.list_param(tags, 'tags', tuple)

        return self._slice(
            [r for r in self.all_runs() if all(r.tags.get(key) == value for key, value in tags)],
            cursor,
            limit,
        )

    def _slice(self, runs, cursor, limit):
        if cursor:
            try:
                index = next(i for i, run in enumerate(runs) if run.run_id == cursor)
            except StopIteration:
                return []
            start = index + 1
        else:
            start = 0

        if limit:
            end = start + limit
        else:
            end = None

        return list(runs)[start:end]

    def get_run_by_id(self, run_id):
        check.str_param(run_id, 'run_id')
        return self._runs.get(run_id)

    def get_run_tags(self):
        result = defaultdict(set)
        for r in self.all_runs():
            for k, v in r.tags:
                result[k].add(v)

        return result.items()

    def get_runs_with_status(self, run_status, cursor=None, limit=None):
        check.inst_param(run_status, 'run_status', PipelineRunStatus)
        matching_runs = list(
            filter(lambda run: run.status == run_status, reversed(self._runs.values()))
        )
        return self._slice(matching_runs, cursor=cursor, limit=limit)

    def has_run(self, run_id):
        check.str_param(run_id, 'run_id')
        return run_id in self._runs

    def delete_run(self, run_id):
        check.str_param(run_id, 'run_id')
        del self._runs[run_id]

    def wipe(self):
        self._runs = OrderedDict()


RunStorageSQLMetadata = db.MetaData()
RunsTable = db.Table(
    'runs',
    RunStorageSQLMetadata,
    db.Column('id', db.Integer, primary_key=True, autoincrement=True),
    db.Column('run_id', db.String(255), unique=True),
    db.Column('pipeline_name', db.String),
    db.Column('status', db.String(63)),
    db.Column('run_body', db.String),
    db.Column('create_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
    db.Column('update_timestamp', db.DateTime, server_default=db.text('CURRENT_TIMESTAMP')),
)

RunTagsTable = db.Table(
    'run_tags',
    RunStorageSQLMetadata,
    db.Column('id', db.Integer, primary_key=True, autoincrement=True),
    db.Column('run_id', None, db.ForeignKey('runs.run_id')),
    db.Column('key', db.String),
    db.Column('value', db.String),
)

create_engine = db.create_engine  # exported


class SQLRunStorage(RunStorage):  # pylint: disable=no-init
    @abstractmethod
    def connect(self):
        ''' context manager yielding a connection '''

    def add_run(self, pipeline_run):
        check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)

        conn = self.connect()
        runs_insert = RunsTable.insert().values(  # pylint: disable=no-value-for-parameter
            run_id=pipeline_run.run_id,
            pipeline_name=pipeline_run.pipeline_name,
            status=pipeline_run.status.value,
            run_body=serialize_dagster_namedtuple(pipeline_run),
        )
        conn.execute(runs_insert)
        if pipeline_run.tags and len(pipeline_run.tags) > 0:
            conn.execute(
                RunTagsTable.insert(),  # pylint: disable=no-value-for-parameter
                [
                    dict(run_id=pipeline_run.run_id, key=k, value=v)
                    for k, v in pipeline_run.tags.items()
                ],
            )

        return pipeline_run

    def handle_run_event(self, run_id, event):
        check.str_param(run_id, 'run_id')
        check.inst_param(event, 'event', DagsterEvent)

        lookup = {
            DagsterEventType.PIPELINE_START: PipelineRunStatus.STARTED,
            DagsterEventType.PIPELINE_SUCCESS: PipelineRunStatus.SUCCESS,
            DagsterEventType.PIPELINE_FAILURE: PipelineRunStatus.FAILURE,
        }

        if event.event_type not in lookup:
            return

        run = self.get_run_by_id(run_id)
        if not run:
            # TODO log?
            return

        new_pipeline_status = lookup[event.event_type]

        self.connect().execute(
            RunsTable.update()  # pylint: disable=no-value-for-parameter
            .where(RunsTable.c.run_id == run_id)
            .values(
                status=new_pipeline_status.value,
                run_body=serialize_dagster_namedtuple(run.run_with_status(new_pipeline_status)),
                update_timestamp=datetime.now(),
            )
        )

    def _rows_to_runs(self, rows):
        return list(map(lambda r: deserialize_json_to_dagster_namedtuple(r[0]), rows))

    def _build_query(self, query, cursor, limit):
        ''' Helper function to deal with cursor/limit pagination args '''

        if cursor:
            cursor_query = db.select([RunsTable.c.id]).where(RunsTable.c.run_id == cursor)
            query = query.where(RunsTable.c.id < cursor_query)

        if limit:
            query = query.limit(limit)

        query = query.order_by(RunsTable.c.id.desc())
        return query

    def all_runs(self, cursor=None, limit=None):
        '''Return all the runs present in the storage.

        Returns:
            List[PipelineRun]: Tuples of run_id, pipeline_run.
        '''
        query = self._build_query(db.select([RunsTable.c.run_body]), cursor, limit)
        rows = self.connect().execute(query).fetchall()
        return self._rows_to_runs(rows)

    def get_runs_with_pipeline_name(self, pipeline_name, cursor=None, limit=None):
        '''Return all the runs present in the storage for a given pipeline.

        Args:
            pipeline_name (str): The pipeline to index on

        Returns:
            List[PipelineRun]: Tuples of run_id, pipeline_run.
        '''
        check.str_param(pipeline_name, 'pipeline_name')

        base_query = db.select([RunsTable.c.run_body]).where(
            RunsTable.c.pipeline_name == pipeline_name
        )
        query = self._build_query(base_query, cursor, limit)
        rows = self.connect().execute(query).fetchall()
        return self._rows_to_runs(rows)

    def get_run_count_with_matching_tags(self, tags):
        sub_query = db.select([1]).select_from(
            RunsTable.outerjoin(RunTagsTable, RunsTable.c.run_id == RunTagsTable.c.run_id)
        )

        sub_query = sub_query.where(
            db.or_(
                *(
                    db.and_(RunTagsTable.c.key == key, RunTagsTable.c.value == value)
                    for key, value in tags
                )
            )
        ).group_by(RunsTable.c.run_id)

        if len(tags) > 0:
            sub_query = sub_query.having(db.func.count(RunsTable.c.run_id) == len(tags))

        sub_query = sub_query.alias("matching_runs")

        query = db.select([db.func.count()]).select_from(sub_query)

        rows = self.connect().execute(query).fetchall()

        count = rows[0][0]
        return count

    def get_runs_with_matching_tags(self, tags, cursor=None, limit=None):
        check.list_param(tags, 'tags', tuple)

        base_query = db.select([RunsTable.c.run_body]).select_from(
            RunsTable.outerjoin(RunTagsTable, RunsTable.c.run_id == RunTagsTable.c.run_id)
        )

        base_query = base_query.where(
            db.or_(
                *(
                    db.and_(RunTagsTable.c.key == key, RunTagsTable.c.value == value)
                    for key, value in tags
                )
            )
        ).group_by(RunsTable.c.run_body, RunsTable.c.id)

        if len(tags) > 0:
            base_query = base_query.having(db.func.count(RunsTable.c.run_id) == len(tags))

        query = self._build_query(base_query, cursor, limit)
        rows = self.connect().execute(query).fetchall()

        return self._rows_to_runs(rows)

    def get_runs_with_status(self, run_status, cursor=None, limit=None):
        check.inst_param(run_status, 'run_status', PipelineRunStatus)

        base_query = db.select([RunsTable.c.run_body]).where(RunsTable.c.status == run_status.value)
        query = self._build_query(base_query, cursor, limit)
        rows = self.connect().execute(query).fetchall()
        return self._rows_to_runs(rows)

    def get_run_by_id(self, run_id):
        '''Get a run by its id.

        Args:
            run_id (str): The id of the run

        Returns:
            Optional[PipelineRun]
        '''
        check.str_param(run_id, 'run_id')

        query = db.select([RunsTable.c.run_body]).where(RunsTable.c.run_id == run_id)
        rows = self.connect().execute(query).fetchall()
        return deserialize_json_to_dagster_namedtuple(rows[0][0]) if len(rows) else None

    def get_run_tags(self):
        result = dict()
        query = db.select([RunTagsTable.c.key, RunTagsTable.c.value]).distinct(RunTagsTable.c.value)
        rows = self.connect().execute(query).fetchall()
        for r in rows:
            if r[0] not in result:
                result[r[0]] = [r[1]]
            else:
                result[r[0]].append(r[1])
        return result.items()

    def has_run(self, run_id):
        check.str_param(run_id, 'run_id')
        return bool(self.get_run_by_id(run_id))

    def delete_run(self, run_id):
        check.str_param(run_id, 'run_id')
        query = db.delete(RunsTable).where(RunsTable.c.run_id == run_id)
        self.connect().execute(query)

    def wipe(self):
        '''Clears the run storage.'''
        conn = self.connect()
        conn.execute(RunsTable.delete())  # pylint: disable=no-value-for-parameter
        conn.execute(RunTagsTable.delete())  # pylint: disable=no-value-for-parameter



from .config_schema import input_hydration_config
from .field_utils import PermissiveDict
from .runtime import RuntimeType, define_python_dagster_type, resolve_to_runtime_type


@input_hydration_config(PermissiveDict())
def _dict_input(_context, value):
    return value


PythonDict = define_python_dagster_type(
    dict,
    'PythonDict',
    input_hydration_config=_dict_input,
    description='''Represents a python dictionary to pass between solids''',
)


def create_typed_runtime_dict(key_dagster_type, value_dagster_type):
    key_type = resolve_to_runtime_type(key_dagster_type)
    value_type = resolve_to_runtime_type(value_dagster_type)

    class _TypedPythonDict(RuntimeType):
        def __init__(self):
            self.key_type = key_type
            self.value_type = value_type
            super(_TypedPythonDict, self).__init__(
                key='TypedPythonDict.{}.{}'.format(key_type.key, value_type.key),
                name=None,
                is_builtin=True,
            )

        def type_check(self, value):
            from dagster.core.definitions.events import Failure

            if not isinstance(value, dict):
                raise Failure('Value {value} should be a dict'.format(value=value))

            for key, value in value.items():
                key_type.type_check(key)
                value_type.type_check(value)

        @property
        def display_name(self):
            return 'Dict[{key},{value}]'.format(
                key=self.key_type.display_name, value=self.value_type.display_name
            )

    return _TypedPythonDict



import glob
import os
import sqlite3
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from contextlib import contextmanager

import gevent.lock
import pyrsistent
import six
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from dagster import check, seven
from dagster.core.definitions.environment_configs import SystemNamedDict
from dagster.core.errors import DagsterError
from dagster.core.events import DagsterEventType
from dagster.core.events.log import EventRecord
from dagster.core.execution.stats import build_stats_from_events
from dagster.core.serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_json_to_dagster_namedtuple,
    serialize_dagster_namedtuple,
)
from dagster.core.types import Field, String
from dagster.utils import mkdir_p

from .pipeline_run import PipelineRunStatsSnapshot, PipelineRunStatus


class EventLogInvalidForRun(DagsterError):
    def __init__(self, *args, **kwargs):
        self.run_id = check.str_param(kwargs.pop('run_id'), 'run_id')
        super(EventLogInvalidForRun, self).__init__(*args, **kwargs)


class EventLogSequence(pyrsistent.CheckedPVector):
    __type__ = EventRecord


class EventLogStorage(six.with_metaclass(ABCMeta)):
    '''Abstract base class for storing structured event logs from pipeline runs.'''

    @abstractmethod
    def get_logs_for_run(self, run_id, cursor=-1):
        '''Get all of the logs corresponding to a run.

        Args:
            run_id (str): The id of the run for which to fetch logs.
            cursor (Optional[int]): Zero-indexed logs will be returned starting from cursor + 1,
                i.e., if cursor is -1, all logs will be returned. (default: -1)
        '''

    def get_stats_for_run(self, run_id):
        '''Get a summary of events that have ocurred in a run.'''

        return build_stats_from_events(run_id, self.get_logs_for_run(run_id))

    @abstractmethod
    def store_event(self, event):
        '''Store an event corresponding to a pipeline run.

        Args:
            run_id (str): The id of the run that generated the event.
            event (EventRecord): The event to store.
        '''

    @abstractmethod
    def delete_events(self, run_id):
        '''Remove events for a given run id'''

    @abstractmethod
    def wipe(self):
        '''Clear the log storage.'''


class WatchableEventLogStorage(EventLogStorage):
    @abstractmethod
    def watch(self, run_id, start_cursor, callback):
        '''Call this method to start watching.'''

    @abstractmethod
    def end_watch(self, run_id, handler):
        '''Call this method to stop watching.'''


class InMemoryEventLogStorage(EventLogStorage):
    def __init__(self):
        self._logs = defaultdict(EventLogSequence)
        self._lock = defaultdict(gevent.lock.Semaphore)

    def get_logs_for_run(self, run_id, cursor=-1):
        check.str_param(run_id, 'run_id')
        check.int_param(cursor, 'cursor')
        check.invariant(
            cursor >= -1,
            'Don\'t know what to do with negative cursor {cursor}'.format(cursor=cursor),
        )

        cursor = cursor + 1
        with self._lock[run_id]:
            return self._logs[run_id][cursor:]

    def store_event(self, event):
        check.inst_param(event, 'event', EventRecord)
        run_id = event.run_id
        with self._lock[run_id]:
            self._logs[run_id] = self._logs[run_id].append(event)

    def delete_events(self, run_id):
        with self._lock[run_id]:
            del self._logs[run_id]
        del self._lock[run_id]

    def wipe(self):
        self._logs = defaultdict(EventLogSequence)
        self._lock = defaultdict(gevent.lock.Semaphore)


CREATE_EVENT_LOG_SQL = '''
CREATE TABLE IF NOT EXISTS event_logs (
    row_id INTEGER PRIMARY KEY AUTOINCREMENT,
    event TEXT NOT NULL,
    dagster_event_type TEXT,
    timestamp TEXT
)
'''

FETCH_EVENTS_SQL = '''
SELECT event FROM event_logs WHERE row_id > ? ORDER BY row_id ASC
'''

FETCH_STATS_SQL = '''
SELECT dagster_event_type, COUNT(1), MAX(timestamp) FROM event_logs GROUP BY dagster_event_type
'''

INSERT_EVENT_SQL = '''
INSERT INTO event_logs (event, dagster_event_type, timestamp) VALUES (?, ?, ?)
'''


class SqliteEventLogStorage(WatchableEventLogStorage, ConfigurableClass):
    def __init__(self, base_dir, inst_data=None):
        '''Note that idempotent initialization of the SQLite database is done on a per-run_id
        basis in the body of store_event, since each run is stored in a separate database.'''
        self._base_dir = check.str_param(base_dir, 'base_dir')
        mkdir_p(self._base_dir)

        self._known_run_ids = set([])
        self._watchers = {}
        self._obs = Observer()
        self._obs.start()
        self._inst_data = check.opt_inst_param(inst_data, 'inst_data', ConfigurableClassData)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return SystemNamedDict('SqliteEventLogStorageConfig', {'base_dir': Field(String)})

    @staticmethod
    def from_config_value(inst_data, config_value, **kwargs):
        return SqliteEventLogStorage(inst_data=inst_data, **dict(config_value, **kwargs))

    @contextmanager
    def _connect(self, run_id):
        try:
            with sqlite3.connect(self.filepath_for_run_id(run_id)) as conn:
                yield conn
        finally:
            conn.close()

    def filepath_for_run_id(self, run_id):
        check.str_param(run_id, 'run_id')
        return os.path.join(self._base_dir, '{run_id}.db'.format(run_id=run_id))

    def store_event(self, event):
        check.inst_param(event, 'event', EventRecord)
        run_id = event.run_id
        if not run_id in self._known_run_ids:
            with self._connect(run_id) as conn:
                conn.cursor().execute(CREATE_EVENT_LOG_SQL)
                conn.cursor().execute('PRAGMA journal_mode=WAL;')
                self._known_run_ids.add(run_id)
        with self._connect(run_id) as conn:
            dagster_event_type = None
            if event.is_dagster_event:
                dagster_event_type = event.dagster_event.event_type_value

            conn.cursor().execute(
                INSERT_EVENT_SQL,
                (serialize_dagster_namedtuple(event), dagster_event_type, event.timestamp),
            )

    def get_logs_for_run(self, run_id, cursor=-1):
        check.str_param(run_id, 'run_id')
        check.int_param(cursor, 'cursor')
        check.invariant(
            cursor >= -1,
            'Don\'t know what to do with negative cursor {cursor}'.format(cursor=cursor),
        )

        events = []
        if not os.path.exists(self.filepath_for_run_id(run_id)):
            return events

        cursor += 1  # adjust from 0 based offset to 1
        try:
            with self._connect(run_id) as conn:
                results = conn.cursor().execute(FETCH_EVENTS_SQL, (str(cursor),)).fetchall()
        except sqlite3.Error as err:
            six.raise_from(EventLogInvalidForRun(run_id=run_id), err)

        try:
            for (json_str,) in results:
                events.append(
                    check.inst_param(
                        deserialize_json_to_dagster_namedtuple(json_str), 'event', EventRecord
                    )
                )
        except (seven.JSONDecodeError, check.CheckError) as err:
            six.raise_from(EventLogInvalidForRun(run_id=run_id), err)

        return events

    def get_stats_for_run(self, run_id):
        if not os.path.exists(self.filepath_for_run_id(run_id)):
            return None

        try:
            with self._connect(run_id) as conn:
                results = conn.cursor().execute(FETCH_STATS_SQL).fetchall()
        except sqlite3.Error as err:
            six.raise_from(EventLogInvalidForRun(run_id=run_id), err)

        try:
            counts = {}
            times = {}
            for result in results:
                if result[0]:
                    counts[result[0]] = result[1]
                    times[result[0]] = result[2]

            return PipelineRunStatsSnapshot(
                run_id=run_id,
                steps_succeeded=counts.get(DagsterEventType.STEP_SUCCESS.value, 0),
                steps_failed=counts.get(DagsterEventType.STEP_FAILURE.value, 0),
                materializations=counts.get(DagsterEventType.STEP_MATERIALIZATION.value, 0),
                expectations=counts.get(DagsterEventType.STEP_EXPECTATION_RESULT.value, 0),
                start_time=float(times.get(DagsterEventType.PIPELINE_START.value, 0.0)),
                end_time=float(
                    times.get(
                        DagsterEventType.PIPELINE_SUCCESS.value,
                        times.get(DagsterEventType.PIPELINE_FAILURE.value, 0.0),
                    )
                ),
            )
        except (seven.JSONDecodeError, check.CheckError) as err:
            six.raise_from(EventLogInvalidForRun(run_id=run_id), err)

    def wipe(self):
        for filename in glob.glob(os.path.join(self._base_dir, '*.db')):
            os.unlink(filename)

    def delete_events(self, run_id):
        path = self.filepath_for_run_id(run_id)
        if os.path.exists(path):
            os.unlink(path)

    @property
    def is_persistent(self):
        return True

    def watch(self, run_id, start_cursor, callback):
        watchdog = EventLogStorageWatchdog(self, run_id, callback, start_cursor)
        self._watchers[run_id] = self._obs.schedule(watchdog, self._base_dir, True)

    def end_watch(self, run_id, handler):
        self._obs.remove_handler_for_watch(handler, self._watchers[run_id])
        del self._watchers[run_id]


class EventLogStorageWatchdog(PatternMatchingEventHandler):
    def __init__(self, event_log_storage, run_id, callback, start_cursor, **kwargs):
        self._event_log_storage = check.inst_param(
            event_log_storage, 'event_log_storage', WatchableEventLogStorage
        )
        self._run_id = check.str_param(run_id, 'run_id')
        self._cb = check.callable_param(callback, 'callback')
        self._log_path = event_log_storage.filepath_for_run_id(run_id)
        self._cursor = start_cursor
        super(EventLogStorageWatchdog, self).__init__(patterns=[self._log_path], **kwargs)

    def _process_log(self):
        events = self._event_log_storage.get_logs_for_run(self._run_id, self._cursor)
        self._cursor += len(events)
        for event in events:
            status = self._cb(event)

            if status == PipelineRunStatus.SUCCESS or status == PipelineRunStatus.FAILURE:
                self._event_log_storage.end_watch(self._run_id, self)

    def on_created(self, event):
        check.invariant(event.src_path == self._log_path)
        self._process_log()

    def on_modified(self, event):
        check.invariant(event.src_path == self._log_path)
        self._process_log()



import collections

import pytest

from dagster import DagsterInvalidDefinitionError, as_dagster_type, dagster_type
from dagster.core.types.runtime import RuntimeType, resolve_to_runtime_type


def test_dagster_type_decorator():
    @dagster_type(name=None)
    class Foo(object):
        pass

    @dagster_type()
    class Bar(object):
        pass

    @dagster_type
    class Baaz(object):
        pass

    assert resolve_to_runtime_type(Foo).name == 'Foo'
    assert resolve_to_runtime_type(Bar).name == 'Bar'
    assert resolve_to_runtime_type(Baaz).name == 'Baaz'


def test_dagster_type_decorator_name_desc():
    @dagster_type(name='DifferentName', description='desc')
    class Something(object):
        pass

    runtime_type = resolve_to_runtime_type(Something)
    assert runtime_type.name == 'DifferentName'
    assert runtime_type.description == 'desc'


def test_make_dagster_type():
    OverwriteNameTuple = as_dagster_type(collections.namedtuple('SomeNamedTuple', 'prop'))
    runtime_type = resolve_to_runtime_type(OverwriteNameTuple)
    assert runtime_type.name == 'SomeNamedTuple'
    assert OverwriteNameTuple(prop='foo').prop == 'foo'

    OverwriteNameTuple = as_dagster_type(
        collections.namedtuple('SomeNamedTuple', 'prop'), name='OverwriteName'
    )
    runtime_type = resolve_to_runtime_type(OverwriteNameTuple)
    assert runtime_type.name == 'OverwriteName'
    assert OverwriteNameTuple(prop='foo').prop == 'foo'


def test_make_dagster_type_from_builtin():
    OrderedDict = as_dagster_type(collections.OrderedDict)
    assert OrderedDict is collections.OrderedDict
    assert OrderedDict([('foo', 'bar')]) == collections.OrderedDict([('foo', 'bar')])
    assert isinstance(resolve_to_runtime_type(OrderedDict), RuntimeType)
    assert resolve_to_runtime_type(OrderedDict).python_type is collections.OrderedDict


def test_dagster_type_collision():
    class Foo:
        pass

    _Foo_1 = as_dagster_type(Foo)
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='A Dagster runtime type has already been registered for the Python type',
    ):
        _Foo_2 = as_dagster_type(Foo)



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError

from .builtin_enum import BuiltinEnum
from .config import Any, ConfigType, List, Nullable, Set, Tuple
from .default_applier import apply_default_values
from .field_utils import FIELD_NO_DEFAULT_PROVIDED, all_optional_type
from .wrapping import WrappingListType, WrappingNullableType, WrappingSetType, WrappingTupleType


def resolve_to_config_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_config_type(list_type.inner_type))


def resolve_to_config_set(set_type):
    check.inst_param(set_type, 'set_type', WrappingSetType)
    return Set(resolve_to_config_type(set_type.inner_type))


def resolve_to_config_tuple(tuple_type):
    check.inst_param(tuple_type, 'list_type', WrappingTupleType)
    if tuple_type.inner_type is None:
        return List(resolve_to_config_type(tuple_type.inner_type))
    inner_types = [resolve_to_config_type(inner_type) for inner_type in tuple_type.inner_type]
    return Tuple(inner_types)


def resolve_to_config_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Nullable(resolve_to_config_type(nullable_type.inner_type))


def resolve_to_config_type(dagster_type):
    from .mapping import remap_python_type
    from .runtime import RuntimeType

    check.invariant(
        not (isinstance(dagster_type, type) and issubclass(dagster_type, RuntimeType)),
        'Cannot resolve a runtime type to a config type',
    )

    dagster_type = remap_python_type(dagster_type)

    if dagster_type is None:
        return Any.inst()
    if BuiltinEnum.contains(dagster_type):
        return ConfigType.from_builtin_enum(dagster_type)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_config_list(dagster_type).inst()
    if isinstance(dagster_type, WrappingSetType):
        return resolve_to_config_set(dagster_type).inst()
    if isinstance(dagster_type, WrappingTupleType):
        return resolve_to_config_tuple(dagster_type).inst()
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_config_nullable(dagster_type).inst()
    if isinstance(dagster_type, type) and issubclass(dagster_type, ConfigType):
        return dagster_type.inst()

    return None


class Field(object):
    '''Defines the schema for a configuration field.
    
    Config fields are parsed according to their schemas in order to yield values available at
    pipeline execution time through the config system. Config fields can be set on solids, on custom
    data types (as the :py:func:`@input_hydration_schema <dagster.input_hydration_schema>`), and on
    other pluggable components of the system, such as resources, loggers, and executors.

    Args:
        dagster_type (Any): The type of this field. Users should provide one of the
            :ref:`built-in types <builtin>`, a composite constructed using :py:func:`Selector`
            or :py:func:`PermissiveDict`, or a dagster type constructed with
            :py:func:`as_dagster_type`, :py:func:`@dagster_type <dagster_type`, or
            :py:func:`define_python_dagster_type` that has an ``input_hydration_config`` defined.
            Note that these constructs can be nested -- i.e., a :py:class:`Dict` can itself contain
            :py:class:`Fields <Field>` of other types, etc.
        default_value (Any): A default value for this field, conformant to the schema set by the
            ``dagster_type`` argument. If a default value is provided, ``is_optional`` should be
            ``True``.
        is_optional (bool): Whether the presence of this field is optional. If ``is_optional`` is
            ``False``, no default value should be provided.
        description (str): A human-readable description of this config field.

    **Examples**:

    .. code-block:: python

        @solid(
            config_field=Field(
                Dict({'word': Field(String, default_value='foo'), 'repeats': Field(Int)})
            )
        )
        def repeat_word(context):
            return context.solid_config['word'] * context.solid_config['repeats']
    '''

    def __init__(
        self,
        dagster_type,
        default_value=FIELD_NO_DEFAULT_PROVIDED,
        is_optional=None,
        is_secret=False,
        description=None,
    ):
        if not isinstance(dagster_type, ConfigType):
            config_type = resolve_to_config_type(dagster_type)
            if not config_type:
                raise DagsterInvalidDefinitionError(
                    (
                        'Attempted to pass {value_repr} to a Field that expects a valid '
                        'dagster type usable in config (e.g. Dict, Int, String et al).'
                    ).format(value_repr=repr(dagster_type))
                )
        else:
            config_type = dagster_type

        self.config_type = check.inst_param(config_type, 'config_type', ConfigType)

        self.description = check.opt_str_param(description, 'description')
        if is_optional is None:
            is_optional = all_optional_type(self.config_type)
            if is_optional is True:
                self._default_value = apply_default_values(self.config_type, None)
            else:
                self._default_value = default_value
        else:
            is_optional = check.bool_param(is_optional, 'is_optional')
            self._default_value = default_value

        if is_optional is False:
            check.param_invariant(
                default_value == FIELD_NO_DEFAULT_PROVIDED,
                'default_value',
                'required arguments should not specify default values',
            )

        if config_type.is_scalar and self._default_value != FIELD_NO_DEFAULT_PROVIDED:
            check.param_invariant(
                config_type.is_config_scalar_valid(self._default_value),
                'default_value',
                'default value not valid for config type {name}, got value {val} of type {type}'.format(
                    name=config_type.name, val=self._default_value, type=type(self._default_value)
                ),
            )

        self.is_optional = is_optional

        self.is_secret = check.bool_param(is_secret, 'is_secret')

    @property
    def is_required(self):
        return not self.is_optional

    @property
    def default_provided(self):
        '''Was a default value provided

        Returns:
            bool: Yes or no
        '''
        return self._default_value != FIELD_NO_DEFAULT_PROVIDED

    @property
    def default_value(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return self._default_value()

        return self._default_value

    @property
    def default_value_as_str(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return repr(self._default_value)

        return str(self._default_value)

    def __repr__(self):
        return (
            'Field({config_type}, default={default}, is_optional={is_optional}, '
            'is_secret={is_secret})'
        ).format(
            config_type=self.config_type,
            default='@'
            if self._default_value == FIELD_NO_DEFAULT_PROVIDED
            else self._default_value,
            is_optional=self.is_optional,
            is_secret=self.is_secret,
        )


def check_field_param(obj, param_name):
    return check.inst_param(obj, param_name, Field)


def check_opt_field_param(obj, param_name):
    return check.opt_inst_param(obj, param_name, Field)



from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.execution.config import IRunConfig
from dagster.core.types import Dict, Field


def resolve_config_field(config_field, config, source):
    if config_field is not None and config is not None:
        raise DagsterInvalidDefinitionError(
            'Must only provide one of config_field or config but not both in {}.'
            'Using the config arg is equivalent to config_field=Field(Dict(...)).'.format(source)
        )

    if config_field:
        return config_field

    if config:
        return Field(Dict(config))

    return None


class ConfigMapping(namedtuple('_ConfigMapping', 'config_fn config_field')):
    '''Defines a config mapping for a composite solid.

    By specifying a config mapping function, you can override the configuration for the child
    solids contained within a composite solid.
            
    Config mappings require the configuration field to be specified as ``config``, which will be
    exposed as the configuration field for the composite solid, as well as a configuration mapping
    function, ``config_fn``, which maps the config provided to the composite solid to the config
    that will be provided to the child solids.
    
    Args:
        config_fn (Callable[[ConfigMappingContext, dict], dict]): The function that will be called
            to map the composite config to a config appropriate for the child solids.
        config_field (Field): The schema of the composite config.
    '''

    def __new__(cls, config_fn, config=None):
        config = check.opt_dict_param(config, 'config')

        return super(ConfigMapping, cls).__new__(
            cls,
            config_fn=check.callable_param(config_fn, 'config_fn'),
            config_field=resolve_config_field(None, config, 'ConfigMapping'),
        )


class ConfigMappingContext(namedtuple('ConfigMappingContext', 'run_config')):
    '''Config mapping-specific context.

    Attributes:
        run_config (RunConfig): The run config belonging to this pipeline run.
    '''

    def __new__(cls, run_config=None):
        return super(ConfigMappingContext, cls).__new__(
            cls, check.inst_param(run_config, 'run_config', IRunConfig)
        )



from enum import Enum as PythonEnum

import pytest

from dagster import (
    DagsterInvalidConfigError,
    Dict,
    Enum,
    EnumValue,
    Field,
    Int,
    PipelineDefinition,
    execute_pipeline,
    solid,
)
from dagster.core.types.config import ConfigEnum
from dagster.core.types.evaluator import evaluate_config


def define_test_enum_type():
    return ConfigEnum(name='TestEnum', enum_values=[EnumValue('VALUE_ONE')])


def test_config_enums():
    assert evaluate_config(define_test_enum_type(), 'VALUE_ONE').success


def test_config_enum_error_none():
    assert not evaluate_config(define_test_enum_type(), None).success


def test_config_enum_error_wrong_type():
    assert not evaluate_config(define_test_enum_type(), 384934).success


def test_config_enum_error():
    assert not evaluate_config(define_test_enum_type(), 'NOT_PRESENT').success


def test_enum_in_pipeline_execution():
    called = {}

    @solid(
        config_field=Field(
            Dict(
                {
                    'int_field': Field(Int),
                    'enum_field': Field(Enum('AnEnum', [EnumValue('ENUM_VALUE')])),
                }
            )
        )
    )
    def config_me(context):
        assert context.solid_config['int_field'] == 2
        assert context.solid_config['enum_field'] == 'ENUM_VALUE'
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='enum_in_pipeline', solid_defs=[config_me])

    result = execute_pipeline(
        pipeline_def,
        {'solids': {'config_me': {'config': {'int_field': 2, 'enum_field': 'ENUM_VALUE'}}}},
    )

    assert result.success
    assert called['yup']

    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        execute_pipeline(
            pipeline_def,
            {'solids': {'config_me': {'config': {'int_field': 2, 'enum_field': 'NOPE'}}}},
        )

    assert 'Error 1: Type failure at path "root:solids:config_me:config:enum_field"' in str(
        exc_info.value
    )


class NativeEnum(PythonEnum):
    FOO = 1
    BAR = 2


def test_native_enum_dagster_enum():
    dagster_enum = Enum(
        'DagsterNativeEnum',
        [
            EnumValue(config_value='FOO', python_value=NativeEnum.FOO),
            EnumValue(config_value='BAR', python_value=NativeEnum.BAR),
        ],
    )

    called = {}

    @solid(config_field=Field(dagster_enum))
    def dagster_enum_me(context):
        assert context.solid_config == NativeEnum.BAR
        called['yup'] = True

    pipeline_def = PipelineDefinition(
        name='native_enum_dagster_pipeline', solid_defs=[dagster_enum_me]
    )

    result = execute_pipeline(pipeline_def, {'solids': {'dagster_enum_me': {'config': 'BAR'}}})
    assert result.success
    assert called['yup']



import typing

from .builtin_enum import BuiltinEnum
from .config_schema import (
    InputHydrationConfig,
    OutputMaterializationConfig,
    input_hydration_config,
    input_selector_schema,
    output_materialization_config,
    output_selector_schema,
)
from .field import Field
from .field_utils import Dict, NamedDict, NamedSelector, PermissiveDict, Selector
from .python_dict import PythonDict
from .python_set import PythonSet
from .python_tuple import PythonTuple
from .runtime import PythonObjectType, define_python_dagster_type
from .wrapping import List, Optional

Any = BuiltinEnum.ANY
String = BuiltinEnum.STRING
Int = BuiltinEnum.INT
Bool = BuiltinEnum.BOOL
Path = BuiltinEnum.PATH
Float = BuiltinEnum.FLOAT
Nothing = BuiltinEnum.NOTHING

Set = typing.Set
Tuple = typing.Tuple

# What sort of witchcraft is this?

# Dagster actually has two type systems, one for *runtime* values, and one
# for schematizing config. However, we wish to avoid exposing the differences
# to users, so here we are. From their standpoint there is just one thing to
# deal with. That is why, for example, there are actually *three* different types
# of Lists in this submodule. One is the config version of List; another is the
# runtime version of List; and the last is the user-facing version of List, that
# -- depending on context -- is converted (resolved, in the parlance) to either
# a config or a runtime type depending on whether it was, say passed into a Field
# (in which case it would be config version) or a InputDefinition (in which case it would be
# a runtime version.)

# This can be confusing, so we follow some naming conventions:

# 1. runtime_cls: Runtime types are singleton instances of a *class*. Users define
# classes and pass those in directly. When you know you are dealing with one of these
# classes, this should be the naming convention
# 2. config_cls: Similar to runtime_cls, but for config types.
# 3. runtime_type: When isinstance(runtime_type, RuntimeType) is guaranteed to be true.
# The general idiom is convert a {config,runtime}_cls to a type instance as soon as possible.
# The internal guts of all these systems operate on type instances for the most part.
# 4. config_type. When isinstance(config_type, ConfigType) is guaranteed to be true.
# 5. dagster_type. dagster_type is used for **user-facing** apis. These can be, very concretely,
# either builtin-primitives such as Dict, Field, Int, String and so forth, or user-defined types.



from __future__ import absolute_import

import copy
import logging
import sys
import traceback
from collections import namedtuple

from dagster import check, seven
from dagster.core.definitions.logger import logger
from dagster.core.log_manager import PYTHON_LOGGING_LEVELS_MAPPING, coerce_valid_log_level
from dagster.core.types.config import Enum, EnumValue

LogLevelEnum = Enum('log_level', list(map(EnumValue, PYTHON_LOGGING_LEVELS_MAPPING.keys())))


class JsonFileHandler(logging.Handler):
    def __init__(self, json_path):
        super(JsonFileHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')

    def emit(self, record):
        try:
            log_dict = copy.copy(record.__dict__)

            # This horrific monstrosity is to maintain backwards compatability
            # with the old behavior of the JsonFileHandler, which the clarify
            # project has a dependency on. It relied on the dagster-defined
            # properties smashing all the properties of the LogRecord object
            # and uploads all of those properties to a redshift table for
            # in order to do analytics on the log

            if 'dagster_meta' in log_dict:
                dagster_meta_dict = log_dict['dagster_meta']
                del log_dict['dagster_meta']
            else:
                dagster_meta_dict = {}

            log_dict.update(dagster_meta_dict)

            with open(self.json_path, 'a') as ff:
                text_line = seven.json.dumps(log_dict)
                ff.write(text_line + '\n')
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('[{}] Error during logging!'.format(self.__class__.__name__))
            logging.exception(str(e))


class StructuredLoggerMessage(
    namedtuple('_StructuredLoggerMessage', 'name message level meta record')
):
    def __new__(cls, name, message, level, meta, record):
        return super(StructuredLoggerMessage, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(message, 'message'),
            coerce_valid_log_level(level),
            check.dict_param(meta, 'meta'),
            check.inst_param(record, 'record', logging.LogRecord),
        )


class JsonEventLoggerHandler(logging.Handler):
    def __init__(self, json_path, construct_event_record):
        super(JsonEventLoggerHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')
        self.construct_event_record = construct_event_record

    def emit(self, record):
        try:
            event_record = self.construct_event_record(record)
            with open(self.json_path, 'a') as ff:
                text_line = seven.json.dumps(event_record.to_dict())
                ff.write(text_line + '\n')

        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('[{}] Error during logging!'.format(self.__class__.__name__))
            logging.exception(str(e))


class StructuredLoggerHandler(logging.Handler):
    def __init__(self, callback):
        super(StructuredLoggerHandler, self).__init__()
        self.callback = check.is_callable(callback, 'callback')

    def emit(self, record):
        try:
            self.callback(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            )
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('[{}] Error during logging!'.format(self.__class__.__name__))
            logging.exception(str(e))


def construct_single_handler_logger(name, level, handler):
    check.str_param(name, 'name')
    check.inst_param(handler, 'handler', logging.Handler)

    level = coerce_valid_log_level(level)

    @logger
    def single_handler_logger(_init_context):
        klass = logging.getLoggerClass()
        logger_ = klass(name, level=level)
        logger_.addHandler(handler)
        handler.setLevel(level)
        return logger_

    return single_handler_logger


def define_structured_logger(name, callback, level):
    check.str_param(name, 'name')
    check.callable_param(callback, 'callback')
    level = coerce_valid_log_level(level)

    return construct_single_handler_logger(name, level, StructuredLoggerHandler(callback))


def define_json_file_logger(name, json_path, level):
    check.str_param(name, 'name')
    check.str_param(json_path, 'json_path')
    level = coerce_valid_log_level(level)

    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    return construct_single_handler_logger(name, level, stream_handler)


def get_stack_trace_array(exception):
    check.inst_param(exception, 'exception', Exception)
    if hasattr(exception, '__traceback__'):
        tb = exception.__traceback__
    else:
        _exc_type, _exc_value, tb = sys.exc_info()
    return traceback.format_tb(tb)


def default_format_string():
    return '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def define_default_formatter():
    return logging.Formatter(default_format_string())



import inspect

from dagster import check

if hasattr(inspect, 'signature'):
    funcsigs = inspect
else:
    import funcsigs


class InvalidDecoratedFunctionInfo:
    TYPES = {'vararg': 1, 'missing_name': 2, 'extra': 3}

    def __init__(self, error_type, param=None, missing_names=None):
        self.error_type = error_type
        self.param = param
        self.missing_names = missing_names


def split_function_parameters(fn, expected_positionals):
    check.callable_param(fn, 'fn')
    check.list_param(expected_positionals, 'expected_positionals', str)
    fn_params = list(funcsigs.signature(fn).parameters.values())
    return fn_params[0 : len(expected_positionals)], fn_params[len(expected_positionals) :]


def validate_decorated_fn_positionals(decorated_fn_positionals, expected_positionals):
    if len(decorated_fn_positionals) < len(expected_positionals):
        return expected_positionals[0]
    for expected_name, actual in zip(expected_positionals, decorated_fn_positionals):
        if expected_name != '*':
            possible_names = {
                '_',
                expected_name,
                '_{expected}'.format(expected=expected_name),
                '{expected}_'.format(expected=expected_name),
            }
            if (
                actual.kind
                not in [
                    funcsigs.Parameter.POSITIONAL_OR_KEYWORD,
                    funcsigs.Parameter.POSITIONAL_ONLY,
                ]
            ) or (actual.name not in possible_names):
                return expected_name


def validate_decorated_fn_non_positionals(input_def_names, decorated_fn_non_positional_params):
    used_inputs = set()
    has_kwargs = False

    for param in decorated_fn_non_positional_params:
        if param.kind == funcsigs.Parameter.VAR_KEYWORD:
            has_kwargs = True
        elif param.kind == funcsigs.Parameter.VAR_POSITIONAL:
            return InvalidDecoratedFunctionInfo(
                error_type=InvalidDecoratedFunctionInfo.TYPES['vararg']
            )

        else:
            if param.name not in input_def_names:
                return InvalidDecoratedFunctionInfo(
                    InvalidDecoratedFunctionInfo.TYPES['missing_name'], param=param.name
                )
            else:
                used_inputs.add(param.name)

    undeclared_inputs = input_def_names - used_inputs
    if not has_kwargs and undeclared_inputs:
        return InvalidDecoratedFunctionInfo(
            InvalidDecoratedFunctionInfo.TYPES['extra'], missing_names=undeclared_inputs
        )



from collections import namedtuple

from dagster import check
from dagster.core.definitions import (
    CompositeSolidDefinition,
    InputDefinition,
    PipelineDefinition,
    Solid,
    SolidDefinition,
    SolidHandle,
    SolidOutputHandle,
    solids_in_topological_order,
)
from dagster.core.definitions.dependency import DependencyStructure
from dagster.core.errors import DagsterExecutionStepNotFoundError, DagsterInvariantViolationError
from dagster.core.execution.config import IRunConfig
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.utils import toposort

from .compute import create_compute_step
from .objects import ExecutionStep, StepInput, StepInputSourceType, StepOutputHandle


class _PlanBuilder:
    '''_PlanBuilder. This is the state that is built up during the execution plan build process.

    steps List[ExecutionStep]: a list of the execution steps that have been created.

    step_output_map Dict[SolidOutputHandle, StepOutputHandle]:  maps logical solid outputs
    (solid_name, output_name) to particular step outputs. This covers the case where a solid maps to
    multiple steps and one wants to be able to attach to the logical output of a solid during
    execution.
    '''

    def __init__(self, pipeline_def, environment_config, run_config):
        self.pipeline_def = check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        self.environment_config = check.inst_param(
            environment_config, 'environment_config', EnvironmentConfig
        )
        self.run_config = check.inst_param(run_config, 'run_config', IRunConfig)
        self.mode_definition = pipeline_def.get_mode_definition(run_config.mode)
        self._steps = dict()
        self.step_output_map = dict()
        self._seen_keys = set()

    @property
    def pipeline_name(self):
        return self.pipeline_def.name

    def add_step(self, step):
        # Keep track of the step keys we've seen so far to ensure we don't add duplicates
        if step.key in self._seen_keys:
            keys = [s.key for s in self._steps]
            check.failed(
                'Duplicated key {key}. Full list seen so far: {key_list}.'.format(
                    key=step.key, key_list=keys
                )
            )
        self._seen_keys.add(step.key)
        self._steps[step.solid_handle.to_string()] = step

    def add_steps(self, steps):
        for step in steps:
            self.add_step(step)

    def get_step_by_handle(self, handle):
        check.inst_param(handle, 'handle', SolidHandle)
        return self._steps[handle.to_string()]

    def get_output_handle(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return self.step_output_map[key]

    def set_output_handle(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        self.step_output_map[key] = val

    def build(self):
        '''Builds the execution plan.
        '''

        # Recursively build the exeuction plan starting at the root pipeline
        self._build_from_sorted_solids(
            solids_in_topological_order(self.pipeline_def), self.pipeline_def.dependency_structure
        )

        # Construct dependency dictionary
        deps = {step.key: set() for step in self._steps.values()}

        for step in self._steps.values():
            for step_input in step.step_inputs:
                deps[step.key].update(step_input.dependency_keys)

        step_dict = {step.key: step for step in self._steps.values()}

        system_storage_def = self.mode_definition.get_system_storage_def(
            self.environment_config.storage.system_storage_name
        )

        previous_run_id = self.run_config.previous_run_id
        step_keys_to_execute = self.run_config.step_keys_to_execute or list(step_dict.keys())

        return ExecutionPlan(
            self.pipeline_def,
            step_dict,
            deps,
            system_storage_def.is_persistent,
            previous_run_id,
            step_keys_to_execute,
        )

    def _build_from_sorted_solids(
        self, solids, dependency_structure, parent_handle=None, parent_step_inputs=None
    ):
        for solid in solids:
            handle = SolidHandle(solid.name, solid.definition.name, parent_handle)

            ### 1. INPUTS
            # Create and add execution plan steps for solid inputs
            step_inputs = []
            for input_name, input_def in solid.definition.input_dict.items():
                step_input = get_step_input(
                    self,
                    solid,
                    input_name,
                    input_def,
                    dependency_structure,
                    handle,
                    parent_step_inputs,
                )

                # If an input with runtime_type "Nothing" doesnt have a value
                # we don't create a StepInput
                if step_input is None:
                    continue

                check.inst_param(step_input, 'step_input', StepInput)
                step_inputs.append(step_input)

            ### 2a. COMPUTE FUNCTION
            # Create and add execution plan step for the solid compute function
            if isinstance(solid.definition, SolidDefinition):
                solid_compute_step = create_compute_step(
                    self.pipeline_name, self.environment_config, solid, step_inputs, handle
                )
                self.add_step(solid_compute_step)

            ### 2b. RECURSE
            # Recurse over the solids contained in an instance of CompositeSolidDefinition
            elif isinstance(solid.definition, CompositeSolidDefinition):
                self._build_from_sorted_solids(
                    solids_in_topological_order(solid.definition),
                    solid.definition.dependency_structure,
                    parent_handle=handle,
                    parent_step_inputs=step_inputs,
                )

            else:
                check.invariant(
                    False,
                    'Unexpected solid type {type} encountered during execution planning'.format(
                        type=type(solid.definition)
                    ),
                )

            ### 3. OUTPUTS
            # Create output handles for solid outputs
            for name, output_def in solid.definition.output_dict.items():
                output_handle = solid.output_handle(name)

                # Punch through layers of composition scope to map to the output of the
                # actual compute step
                resolved_output_def, resolved_handle = solid.definition.resolve_output_to_origin(
                    output_def.name, handle
                )
                compute_step = self.get_step_by_handle(resolved_handle)
                self.set_output_handle(
                    output_handle,
                    StepOutputHandle.from_step(compute_step, resolved_output_def.name),
                )


def get_step_input(
    plan_builder, solid, input_name, input_def, dependency_structure, handle, parent_step_inputs
):
    check.inst_param(plan_builder, 'plan_builder', _PlanBuilder)
    check.inst_param(solid, 'solid', Solid)
    check.str_param(input_name, 'input_name')
    check.inst_param(input_def, 'input_def', InputDefinition)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)
    check.opt_inst_param(handle, 'handle', SolidHandle)
    check.opt_list_param(parent_step_inputs, 'parent_step_inputs', of_type=StepInput)

    solid_config = plan_builder.environment_config.solids.get(str(handle))
    if solid_config and input_name in solid_config.inputs:
        return StepInput(
            input_name,
            input_def.runtime_type,
            StepInputSourceType.CONFIG,
            config_data=solid_config.inputs[input_name],
        )

    input_handle = solid.input_handle(input_name)
    if dependency_structure.has_singular_dep(input_handle):
        solid_output_handle = dependency_structure.get_singular_dep(input_handle)
        return StepInput(
            input_name,
            input_def.runtime_type,
            StepInputSourceType.SINGLE_OUTPUT,
            [plan_builder.get_output_handle(solid_output_handle)],
        )

    if dependency_structure.has_multi_deps(input_handle):
        solid_output_handles = dependency_structure.get_multi_deps(input_handle)
        return StepInput(
            input_name,
            input_def.runtime_type,
            StepInputSourceType.MULTIPLE_OUTPUTS,
            [
                plan_builder.get_output_handle(solid_output_handle)
                for solid_output_handle in solid_output_handles
            ],
        )

    if solid.container_maps_input(input_name):
        parent_name = solid.container_mapped_input(input_name).definition.name
        parent_inputs = {step_input.name: step_input for step_input in parent_step_inputs}
        if parent_name in parent_inputs:
            parent_input = parent_inputs[parent_name]
            return StepInput(
                input_name,
                input_def.runtime_type,
                parent_input.source_type,
                parent_input.source_handles,
                parent_input.config_data,
            )

    # At this point we have an input that is not hooked up to
    # the output of another solid or provided via environment config.

    # We will allow this for "Nothing" type inputs and continue.
    if input_def.runtime_type.is_nothing:
        return None

    # Otherwise we throw an error.
    raise DagsterInvariantViolationError(
        (
            'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
            'must get a value either (a) from a dependency or (b) from the '
            'inputs section of its configuration.'
        ).format(
            pipeline_name=plan_builder.pipeline_name, solid_name=solid.name, input_name=input_name
        )
    )


class ExecutionPlan(
    namedtuple(
        '_ExecutionPlan',
        'pipeline_def step_dict deps steps artifacts_persisted previous_run_id step_keys_to_execute',
    )
):
    def __new__(
        cls,
        pipeline_def,
        step_dict,
        deps,
        artifacts_persisted,
        previous_run_id,
        step_keys_to_execute,
    ):
        missing_steps = [step_key for step_key in step_keys_to_execute if step_key not in step_dict]
        if missing_steps:
            raise DagsterExecutionStepNotFoundError(
                'Execution plan does not contain step{plural}: {steps}'.format(
                    plural='s' if len(missing_steps) > 1 else '', steps=', '.join(missing_steps)
                ),
                step_keys=missing_steps,
            )

        return super(ExecutionPlan, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            step_dict=check.dict_param(
                step_dict, 'step_dict', key_type=str, value_type=ExecutionStep
            ),
            deps=check.dict_param(deps, 'deps', key_type=str, value_type=set),
            steps=list(step_dict.values()),
            artifacts_persisted=check.bool_param(artifacts_persisted, 'artifacts_persisted'),
            previous_run_id=check.opt_str_param(previous_run_id, 'previous_run_id'),
            step_keys_to_execute=check.list_param(
                step_keys_to_execute, 'step_keys_to_execute', of_type=str
            ),
        )

    def get_step_output(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        step = self.get_step_by_key(step_output_handle.step_key)
        return step.step_output_named(step_output_handle.output_name)

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self.step_dict

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self.step_dict[key]

    def topological_steps(self):
        return [step for step_level in self.topological_step_levels() for step in step_level]

    def topological_step_levels(self):
        return [
            [self.step_dict[step_key] for step_key in sorted(step_key_level)]
            for step_key_level in toposort(self.deps)
        ]

    def execution_steps(self):
        return [step for step_level in self.execution_step_levels() for step in step_level]

    def execution_step_levels(self):
        return [
            [self.step_dict[step_key] for step_key in sorted(step_key_level)]
            for step_key_level in toposort(self._execution_deps())
        ]

    def missing_steps(self):
        return [step_key for step_key in self.step_keys_to_execute if not self.has_step(step_key)]

    def _execution_deps(self):
        step_dict = {k: v for k, v in self.step_dict.items() if k in self.step_keys_to_execute}
        deps = {step.key: set() for step in step_dict.values()}
        for step in step_dict.values():
            for step_input in step.step_inputs:
                deps[step.key].update(
                    step_input.dependency_keys.intersection(self.step_keys_to_execute)
                )
        return deps

    def build_subset_plan(self, step_keys_to_execute):
        check.list_param(step_keys_to_execute, 'step_keys_to_execute', of_type=str)
        return ExecutionPlan(
            self.pipeline_def,
            self.step_dict,
            self.deps,
            self.artifacts_persisted,
            self.previous_run_id,
            step_keys_to_execute,
        )

    @staticmethod
    def build(pipeline_def, environment_config, run_config):
        '''Here we build a new ExecutionPlan from a pipeline definition and the environment config.

        To do this, we iterate through the pipeline's solids in topological order, and hand off the
        execution steps for each solid to a companion _PlanBuilder object.

        Once we've processed the entire pipeline, we invoke _PlanBuilder.build() to construct the
        ExecutionPlan object.
        '''
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
        check.inst_param(run_config, 'run_config', IRunConfig)

        plan_builder = _PlanBuilder(pipeline_def, environment_config, run_config)

        # Finally, we build and return the execution plan
        return plan_builder.build()



from collections import namedtuple

from dagster import check
from dagster.core.definitions import ExecutorDefinition, ModeDefinition, PipelineDefinition
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.system_config.objects import EnvironmentConfig


class InitExecutorContext(
    namedtuple(
        'InitExecutorContext',
        'pipeline_def mode_def executor_def pipeline_run environment_config executor_config',
    )
):
    '''Executor-specific initialization context.

    Attributes:
        pipeline_def (PipelineDefinition): The pipeline definition in scope for execution.
        mode_def (ModeDefinition): The mode in which the pipeline is to be executed.
        executor_def (ExecutorDefinition): The definition of the executor currently being
            constructed.
        run_config (RunConfig): Configuration for this pipeline run.
        environment_config (EnvironmentConfig): The parsed environment configuration for this
            pipeline run.
        executor_config (dict): The parsed config passed to the executor.
    '''

    def __new__(
        cls, pipeline_def, mode_def, executor_def, pipeline_run, environment_config, executor_config
    ):
        return super(InitExecutorContext, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            mode_def=check.inst_param(mode_def, 'mode_def', ModeDefinition),
            executor_def=check.inst_param(executor_def, 'executor_def', ExecutorDefinition),
            pipeline_run=check.inst_param(pipeline_run, 'pipeline_run', PipelineRun),
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            executor_config=check.dict_param(executor_config, executor_config, key_type=str),
        )



import argparse
import sys

from setuptools import find_packages, setup


def long_description():
    return '''
## Dagster
Dagster is a system for building modern data applications.

Combining an elegant programming model and beautiful tools, Dagster allows infrastructure engineers,
data engineers, and data scientists to seamlessly collaborate to process and produce the trusted,
reliable data needed in today's world.
'''.strip()


def get_version(name):
    version = {}
    with open('dagster/version.py') as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    if name == 'dagster':
        return version['__version__']
    elif name == 'dagster-nightly':
        return version['__nightly__']
    else:
        raise Exception('Shouldn\'t be here: bad package name {name}'.format(name=name))


parser = argparse.ArgumentParser()
parser.add_argument('--nightly', action='store_true')


def _do_setup(name='dagster'):
    setup(
        name=name,
        version=get_version(name),
        author='Elementl',
        license='Apache-2.0',
        description='Dagster is an opinionated programming model for data pipelines.',
        long_description=long_description(),
        long_description_content_type='text/markdown',
        url='https://github.com/dagster-io/dagster',
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(exclude=['dagster_tests']),
        install_requires=[
            # standard python 2/3 compatability things
            'enum-compat>=0.0.1',
            'future',
            'funcsigs',
            'contextlib2>=0.5.4',
            'pathlib2>=2.3.4; python_version<"3"',
            # cli
            'click>=5.0',
            'coloredlogs>=6.1',
            'graphviz>=0.8.4',
            # pyyaml pinned for compatibility with docker-compose
            # https://github.com/docker/compose/blob/master/setup.py#L35
            'PyYAML>=3.10,<5',
            # core (not explicitly expressed atm)
            'gevent==1.3.7',
            'pyrsistent>=0.14.8',
            'rx==1.6.1',
            'six>=1.11.0',
            'sqlalchemy>=1.0',
            'typing; python_version<"3"',
            'backports.tempfile; python_version<"3"',
            'toposort>=1.0',
            'watchdog>=0.8.3',
            'psutil >= 1.0; platform_system=="Windows"',
            # https://github.com/mhammond/pywin32/issues/1439
            'pywin32 != 226; platform_system=="Windows"',
        ],
        tests_require=['mock'],
        entry_points={'console_scripts': ['dagster = dagster.cli:main']},
    )


if __name__ == '__main__':
    parsed, unparsed = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + unparsed
    if parsed.nightly:
        _do_setup('dagster-nightly')
    else:
        _do_setup('dagster')



from dagster import check
from dagster.core.decorator_utils import (
    split_function_parameters,
    validate_decorated_fn_positionals,
)
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.utils import ensure_single_item

from .builtin_enum import BuiltinEnum
from .config import ConfigType, List, Nullable
from .wrapping import WrappingListType, WrappingNullableType


class InputHydrationConfig:
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def construct_from_config_value(self, _context, config_value):
        '''
        How to create a runtime value from config data.
        '''
        return config_value


def resolve_config_cls_arg(config_cls):
    if BuiltinEnum.contains(config_cls):
        return ConfigType.from_builtin_enum(config_cls)
    elif isinstance(config_cls, WrappingListType):
        return List(resolve_config_cls_arg(config_cls.inner_type))
    elif isinstance(config_cls, WrappingNullableType):
        return Nullable(resolve_config_cls_arg(config_cls.inner_type))
    else:
        check.type_param(config_cls, 'config_cls')
        check.param_invariant(issubclass(config_cls, ConfigType), 'config_cls')
        return config_cls.inst()


def make_bare_input_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)

    class _InputSchema(InputHydrationConfig):
        @property
        def schema_type(self):
            return config_type

    return _InputSchema()


class OutputMaterializationConfig:
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def materialize_runtime_value(self, _context, _config_value, _runtime_value):
        '''
        How to materialize a runtime value given configuration.
        '''
        check.not_implemented('Must implement')


def _create_input_schema(config_type, func):
    class _InputSchema(InputHydrationConfig):
        @property
        def schema_type(self):
            return config_type

        def construct_from_config_value(self, context, config_value):
            return func(context, config_value)

    return _InputSchema()


def input_hydration_config(config_cls):
    '''Create an input hydration config that maps config data to a runtime value.

    The decorated function should take the execution context and parsed config value and return the
    appropriate runtime value.

    Args:
        config_cls (Any): The type of the config data expected by the decorated function. Users
            should provide one of the :ref:`built-in types <builtin>`, or a composite constructed
            using :py:func:`Selector` or :py:func:`PermissiveDict`.
    
    **Examples**:

    .. code-block:: python

        @input_hydration_config(PermissiveDict())
        def _dict_input(_context, value):
            return value
    '''
    config_type = resolve_config_cls_arg(config_cls)
    EXPECTED_POSITIONALS = ['context', '*']

    def wrapper(func):
        fn_positionals, _ = split_function_parameters(func, EXPECTED_POSITIONALS)
        missing_positional = validate_decorated_fn_positionals(fn_positionals, EXPECTED_POSITIONALS)
        if missing_positional:
            raise DagsterInvalidDefinitionError(
                "@input_hydration_config '{solid_name}' decorated function does not have required positional "
                "parameter '{missing_param}'. Solid functions should only have keyword arguments "
                "that match input names and a first positional parameter named 'context'.".format(
                    solid_name=func.__name__, missing_param=missing_positional
                )
            )
        return _create_input_schema(config_type, func)

    return wrapper


def input_selector_schema(config_cls):
    '''
    A decorator for annotating a function that can take the selected properties
    from a ``config_value`` in to an instance of a custom type.

    Args:
        config_cls (Selector)
    '''
    config_type = resolve_config_cls_arg(config_cls)
    check.param_invariant(config_type.is_selector, 'config_cls')

    def _wrap(func):
        def _selector(context, config_value):
            selector_key, selector_value = ensure_single_item(config_value)
            return func(context, selector_key, selector_value)

        return _create_input_schema(config_type, _selector)

    return _wrap


def _create_output_schema(config_type, func):
    class _OutputSchema(OutputMaterializationConfig):
        @property
        def schema_type(self):
            return config_type

        def materialize_runtime_value(self, context, config_value, runtime_value):
            return func(context, config_value, runtime_value)

    return _OutputSchema()


def output_materialization_config(config_cls):
    '''Create an output materialization hydration config that configurably materializes a runtime
    value.

    The decorated function should take the execution context, the parsed config value, and the
    runtime value and the parsed config data, should materialize the runtime value, and should
    return an appropriate :py:class:`Materialization`.

    Args:
        config_cls (Any): The type of the config data expected by the decorated function. Users
            should provide one of the :ref:`built-in types <builtin>`, or a composite constructed
            using :py:func:`Selector` or :py:func:`PermissiveDict`.

    **Examples**:

    .. code-block:: python

        # Takes a list of dicts such as might be read in using csv.DictReader, as well as a config
        value, and writes 
        @output_materialization_config(Path)
        def df_output_schema(_context, path, value):
            with open(path, 'w') as fd:
                writer = csv.DictWriter(fd, fieldnames=value[0].keys())
                writer.writeheader()
                writer.writerows(rowdicts=value)

            return Materialization.file(path)

    '''
    config_type = resolve_config_cls_arg(config_cls)
    return lambda func: _create_output_schema(config_type, func)


def output_selector_schema(config_cls):
    '''
    A decorator for a annotating a function that can take the selected properties
    of a ``config_value`` and an instance of a custom type and materialize it.

    Args:
        config_cls (Selector):
    '''
    config_type = resolve_config_cls_arg(config_cls)
    check.param_invariant(config_type.is_selector, 'config_cls')

    def _wrap(func):
        def _selector(context, config_value, runtime_value):
            selector_key, selector_value = ensure_single_item(config_value)
            return func(context, selector_key, selector_value, runtime_value)

        return _create_output_schema(config_type, _selector)

    return _wrap
